"""
Datasource: MySQL: Querying

TODO:

- auto_commit solves our flakiness problem between connections (switching around), but it seems to be a problem where we arent calling COMMIT, but we havent found this yet.  Troubleshoot later.  Will turn off auto_commit, and then try changing something, if it flips back and forth randomly over many changes, because its switching connections that were COMMITed, thats the problem.
"""


import threading
import logging

try:
  import mysql.connector
  from mysql.connector import errorcode
  MYSQL = 'ORACLE'
  
except ImportError, e:
  import pymysql
  import pymysql.cursors
  MYSQL = 'PUREPYTHON'

from schemaman.utility.log import Log


# Default connection pool size.  Override with connection_data
DEFAULT_CONNECTION_POOL_SIZE = 20


# Connection Pool of Pools (servers on first level (list), pool of connections on second (Connection objects)): dict of lists of Connection objects
CONNECTION_POOL_POOL = {}
CONNECTION_POOL_POOL_LOCK = threading.Lock()


# Default MySQL connection character set
# DEFAULT_CHARSET = 'utf8'
DEFAULT_CHARSET = 'latin1'


# For every request, they can only take 1 request at a time, so they dont collide, so we lock to ensure they are sequential
REQUEST_LOCKING = True
# REQUEST_LOCKING = False
REQUEST_QUERY_LOCK = {}


# Do we force all queries to hold a mutex for querying?  If our library has thread saftey issue (crashes), we need this turned on
SINGLE_THREADED_DEFAULT = False
# SINGLE_THREADED_DEFAULT = True
SINGLE_THREADED_LOCK = threading.Lock()


#TODO(g): Make this a base class, that each Handler type sub-classes.  Useful in this case, as it's an interface, and some methods are more virtual than others.  It's good to have a base class for the interface, otherwise every handler implements its own base, and they seem more disconnected...
class Connection:
  """This wraps MySQL connection and cursor objects, as well as tracks the progress of any requests, and if it is available for use by a new request."""
  
  def __init__(self, connection_data, server_id, request, is_single_threaded=SINGLE_THREADED_DEFAULT):
    Log('Creating new connection: MySQL: %s' % connection_data['datasource']['database'])
    
    # We need these to actually connect
    self.connection_data = connection_data
    self.server_id = server_id
    self.is_single_threaded = is_single_threaded
    
    # This tells us who is connection.  Release() to make this connection available for other requests.
    self.request = request
    self.request_lock = threading.Lock()
    
    # Generate the server key, since this specifies which CONNECTION_POOL_POOL we are in
    self.server_key = GetServerKey(request)
    
    self.connection = None
    self.cursor = None
    
    # Connect
    self.Connect()
  
  
  def __del__(self):
    try:
      print 'Closing Connection: DEL: %s: %s' % (self, self.Close)
      self.Close()
    
    # Ignore this failure, this happens when the context is lost, because the program is closing, and we are working with NoneTypes, instead of expected types
    except AttributeError, e:
      pass


  def Close(self):
    """Close the cursor and connection, if they are open, and set them to None"""
    Log('Closing connection: MySQL: %s: %s' % (self.connection_data['datasource']['database'], self.server_key))
    
    if self.cursor:
      try:
        self.cursor.close()
      finally:
        self.cursor = None
      
    if self.connection:
      try:
        self.connection.close()
      finally:
        self.connection = None


  def Release(self):
    """Release this connection."""
    Log('Releasing connection: MySQL: %s: %s' % (self.server_key, self.request.username))
    
    if self.request_lock.locked and self.request == None:
      print '\n\nERROR: Request Connection was not locked, but had a request: %s' % self.request
    
    self.request = None
    
    self.request_lock.release()


  def Acquire(self, request):
    """Acquire this Connection for this Request."""
    if self.request_lock.locked():
      raise Exception('Attempting to Acquire a Connection when it is already locked: %s' % request)
    
    # Get the lock
    self.request_lock.acquire()
    
    Log('Acquiring connection: MySQL: %s: %s  (auto_commit=%s)' % (self.server_key, request.username, request.auto_commit))
    self.request = request
    
    # If any transactions werent committed, we obviously dont want them to be, or whatever, theyre gone!
    self.connection.rollback()
    
    # Set the auto-commit based on the request specification
    self.connection.autocommit(self.request.auto_commit)
    

  
  def IsAvailable(self):
    """Returns boolean, True if not currently being used by a request and has a non-None connection.
    
    NOTE(g): This does not verify the connection, just ensures that we think we have a valid connection.
    """
    is_locked = self.request_lock.locked()
    is_available = not is_locked
    
    return is_available
  

  def IsInUse(self):
    """Returns boolean, True if not currently being used by a request."""
    return not self.IsAvailable()


  def IsUsedByRequest(self, request):
    """Returns boolean, True if currently being used by the same request that is being passed in as request."""
    if self.request == request:
      return True
    
    else:
      return False
  
  
  def GetServerData(self):
    """Returns a dict for server data, which is layered from the server configs"""
    server_data = {}
    
    # Find the master host, which we will assume we are connecting to for now
    found_server = None
    for server_data in self.connection_data['datasource']['servers']:
      if server_data['id'] == self.server_id:
        found_server = server_data
        break
    
    if not found_server:
      raise Exception('Couldnt find ourselves in our server data, something is wrong: %s: %s' % (self.connection_data['alias'], self.server_id))
    
    # Set specific datasource fields, these fields may not be present, and are overridden by the server specific fields
    server_data['database'] = self.connection_data['datasource'].get('database', None)
    server_data['user'] = self.connection_data['datasource'].get('user', None)
    server_data['password_path'] = self.connection_data['datasource'].get('password_path', None)
    
    # Override all previous fields with server specific fields, which gives us the final data (most specific)
    server_data.update(found_server)
    
    return server_data
  
  
  def Connect(self):
    """Connect to the database"""
    server = self.GetServerData()
    
    Log('Connecting to MySQL server: %s: %s' % (server['host'], server['database']))
    
    # Read the password from the first line of the password file
    try:
      password = open(server['password_path']).read().split('\n', 1)[0]
      Log('Loaded password: %s' % server['password_path'])
      
    except Exception, e:
      Log('ERROR: Failed to read from password file: %s' % server['password_path'], logging.ERROR)
      password = None
    
    if MYSQL == 'ORACLE':
      self.connection = mysql.connector.connect(user=server['user'], password=password, host=server['host'], port=server['port'], database=server['database'], use_unicode=True, charset=DEFAULT_CHARSET)
      self.cursor = self.connection.cursor(dictionary=True)
    
    else:
      self.connection = pymysql.Connection(user=server['user'], passwd=password, host=server['host'], port=server['port'], db=server['database'], cursorclass=pymysql.cursors.DictCursor)
      self.cursor = self.connection.cursor()


  def __QueryLock(self):
    """Lock a query"""
    set_request_lock = False
    set_single_threaded_lock = False
    
    if self.is_single_threaded:
      # print 'Aquiring Single Thread Lock: Start'
      SINGLE_THREADED_LOCK.acquire()
      set_single_threaded_lock = True
      # print 'Aquiring Single Thread Lock: Success'
      
    
    if REQUEST_LOCKING and self.request:
      # Ensure any requests we look at, have a lock we can grab
      if self.request.request_number not in REQUEST_QUERY_LOCK:
        REQUEST_QUERY_LOCK[self.request.request_number] = threading.Lock()
      
      # Get the lock
      # print 'Aquiring Request Lock: Start'
      REQUEST_QUERY_LOCK[self.request.request_number].acquire()
      set_request_lock = True
      # print 'Aquiring Request Lock: Success'
    
    return (set_request_lock, set_single_threaded_lock)


  def __QueryUnlock(self, set_request_lock, set_single_threaded_lock):
    """Unlock a query"""
    # print 'Release Query Locks'
    
    # If we set a request lock, release it
    if set_request_lock:
      # try:
      # print 'Releasing Request Lock: Start'
      REQUEST_QUERY_LOCK[self.request.request_number].release()
      # print 'Releasing Request Lock: Success'
      # except Exception, e:
      #   pass
    
    # If we set the single threaded lock
    if set_single_threaded_lock:
      # try:
      # print 'Releasing Single Lock: Start'
      SINGLE_THREADED_LOCK.release()
      # print 'Releasing Single Lock: Success'
      # except Exception, e:
      #   pass


  def Query(self, sql, params=None, commit=True):
    """Query the database via our connection."""
    set_request_lock = None
    set_single_threaded_lock = None
    
    try:
      (set_request_lock, set_single_threaded_lock) = self.__QueryLock()
      
      done = False
      retry = 0
      
      # Loop through errors and stuff, if they are recoverable (connection failures->reconnect)
      while not done:
        try:
          if not params:
            Log('Query: %s' % sql)
          else:
            Log('Query: %s -- %s' % (sql, params))
          
          result = Query(self.connection, self.cursor, sql, params=params, commit=commit)
          done = True
        
        # Handle DB connection problems
        except pymysql.OperationalError, e:
          print 'MySQL: OperationError: %s' % e
          
          retry += 1
          if retry >= 3:
            self.__QueryUnlock(set_request_lock, set_single_threaded_lock)
            raise Exception('Failed %s times: %s' % (retry-1, e))
          
          # Else, reconnect, something went wrong, this is the best way to fix it
          else:
            self.__QueryUnlock(set_request_lock, set_single_threaded_lock)
            self.Connect()
      
    
    # If we are single threaded, release the lock
    finally:
      # If we got through the initial lock (it will be a bool, not None)
      self.__QueryUnlock(set_request_lock, set_single_threaded_lock)
    
    return result
  
  
  def Commit(self):
    """Commit a transaction in flight."""
    result = self.connection.commit()
    
    return result
  
  
  def AbandonCommit(self):
    """Abandon Commit a transaction in flight."""
    result = self.connection.rollback()
    
    return result


def GetServerKey(request):
  """Returns string, the server key, for use in the CONNECTION_POOL_POOL at top level dict"""
  # Generate the server key, since this specifies which CONNECTION_POOL_POOL we are in
  server_key = '%s.%s' % (request.connection_data['alias'], request.server_id)
  
  return server_key
  

def MySQLReleaseConnections(request):
  """Release any connections tied with this request_number
  
  TODO(g): 
  """
  global CONNECTION_POOL_POOL
  
  # Generate the server key, since this specifies which CONNECTION_POOL_POOL we are in
  server_key = GetServerKey(request)

  # Look through the current connection pool, to see if we already have a connection for this request_number
  if server_key in CONNECTION_POOL_POOL:
    for connection in CONNECTION_POOL_POOL[server_key]:
      # If this connection is for the same request, release it
      if connection.IsUsedByRequest(request):
        connection.Release()


def GetConnection(request, server_id=None):
  """Returns a connection to the specified database server_id, based on the request number (may already have a connection for that request)."""
  global CONNECTION_POOL_POOL
  
  # If we didnt have a server_id specified, use the master_server_id
  if server_id != None:
    server_id = server_id
  elif request.server_id == None:
    server_id = request.connection_data['datasource']['master_server_id']
  else:
    server_id = request.server_id
  
  
  # Find the master host, which we will assume we are connecting to for now
  found_server = None
  for server_data in request.connection_data['datasource']['servers']:
    if server_data['id'] == server_id:
      found_server = server_data
      break


  # Generate the server key, since this specifies which CONNECTION_POOL_POOL we are in
  server_key = GetServerKey(request)


  # Look through the current connection pool, to see if we already have a connection for this request_number
  if server_key in CONNECTION_POOL_POOL:
    for connection in CONNECTION_POOL_POOL[server_key]:
      # If this connection is for the same request, use it
      if connection.IsUsedByRequest(request):
        return connection
  

  # Look through current connection pool, to see if we have any available connections in this server, that we can use
  if server_key in CONNECTION_POOL_POOL:
    for connection in CONNECTION_POOL_POOL[server_key]:
      # If this connection is available (not being used in a request)
      if connection.IsAvailable():
        # This request has now acquired this connection
        connection.Acquire(request)
        return connection

  
  # Create the connection
  connection = Connection(request.connection_data, server_id, request)
  connection.Acquire(request)
  
  # Ensure we have a pool for this server in our connection pool pools
  if connection.server_key not in CONNECTION_POOL_POOL:
    try:
      CONNECTION_POOL_POOL_LOCK.acquire()
      Log('Creating new MySQL connection pool: %s' % connection.server_key)
      CONNECTION_POOL_POOL[connection.server_key] = [connection]
    finally:
      CONNECTION_POOL_POOL_LOCK.release()
  
  # Else, add it to the existing connection pool pool
  else:
    try:
      CONNECTION_POOL_POOL_LOCK.acquire()
      Log('Appending to existing MySQL connection pool: %s  (Count: %s)' % (connection.server_key, len(CONNECTION_POOL_POOL[connection.server_key])))
      CONNECTION_POOL_POOL[connection.server_key].append(connection)
    finally:
      CONNECTION_POOL_POOL_LOCK.release()
  
  return connection


def Query(conn, cursor, sql, params=None, commit=True):
  """Query"""
  cursor.execute(sql, params)
  
  if sql.upper().startswith('INSERT'):
    result = cursor.lastrowid
    
    # If we should immediately commit this as a 1 query transaction
    if commit:
      conn.commit()
      
  elif sql.upper().startswith('UPDATE') or sql.upper().startswith('DELETE'):
    # If we should immediately commit this as a 1 query transaction
    if commit:
      conn.commit()
    
    result = None
  
  elif sql.upper().startswith('SELECT') or sql.upper().startswith('SHOW') or sql.upper().startswith('DESC'):
    result = cursor.fetchall()
  
  else:
    result = None
  
  # We completed our work, and are done
  done = True
  
  return result


def SanitizeSQL(text):
  if text == None:
    text = 'NULL'
  else:
    try:
      text = str(text)
      
    except UnicodeEncodeError, e:
      text = str(text.decode('ascii', 'ignore'))
  
  return text.replace("'", "''").replace('\\', '\\\\')




"""
Actions: Config: Version and Change Management

Configure all the version and change management 
"""



# SchemaMan libraries
import schemaman.utility as utility
from schemaman.utility.interactive_input import *
import schemaman.datasource as datasource


# This action's command on the CLI and also in the connection_data.actions dict as a key for our data
ACTION = 'config__version_change_management'


def Action(connection_data, action_input_args):
  """Perform action: Configure Version and Change Management"""
  print 'Configure Version and Change Management'
  
  data = CollectData(connection_data, action_input_args)
  
  # If we got invalid data or aborted, quit this action
  if data == None:
    return '\nAborting configuration...'
  
  return 'Configured Version Management'


def CollectData(connection_data, action_input_args):
  """Collect data that we need, if we dont already have it stored in our connection_data.actions dict"""
  data = {}
  
  # If we have data in our connection_data, use it
  if ACTION in connection_data['actions']:
    #TODO(g): Validate that this data is correct
    data = connection_data['actions'][ACTION]
  
    print '\nThese are the tables you specified:'
    print '  - Version: Working: %s' % data['table_version_working']
    print '  - Version: Change List: %s' % data['table_version_change_list']
    print '  - Version: Commit: %s' % data['table_version_commit']
    print
    print 'Is this correct?  (Enter Yes to use these values, anything else to enter new values)'
    
    confirmation = ReadLine('$ ')
    
    # Use the saved data, otherwise collect new data
    if confirmation.lower() == 'yes':
      return data
    else:
      data = {}
  
  
  # Get the Database table name
  intro_string = 'Name of table to put the Version Working changes in.'
  GetInputField(data, 'table_version_working', 'Version Working Table', 'working table', intro_string, force_strip=True)
  
  # Get the Table table name
  intro_string = 'Name of table to put the Version Pending Change Lists in.'
  GetInputField(data, 'table_version_change_list', 'Version Pending Change List Table', 'change list table', intro_string, force_strip=True)
  
  # Get the Field table name
  intro_string = 'Name of table to put the Version Committed changes in.'
  GetInputField(data, 'table_version_commit', 'Version Commit Table', 'committed table', intro_string, force_strip=True)

  
  print '\nThese are the tables you specified:'
  print '  - Version: Working: %s' % data['table_version_working']
  print '  - Version: Change List: %s' % data['table_version_change_list']
  print '  - Version: Commit: %s' % data['table_version_commit']
  print
  print 'Is this correct?  (Enter Save or this action will abort)'
  
  confirmation = ReadLine('$ ')
  
  # Aborting
  if confirmation.lower() not in ('save',):
    return None
  
  
  # Save our connection data with the updated information
  if confirmation.lower() == 'save':
    connection_data['actions'][ACTION] = data
    
    datasource.SaveConnectionSpec(connection_data)
  
  
  return data
  
  


#!/usr/bin/env python
"""
SchemaMan - Schema Manager.  Cross database schema revision control, data migration and data access.

- Manage changes in the schema versions, allowing publication of schemas with your code that is not bound to a single database
    software package or platform.
- Use as a cross-database method of interacting with data (insert/update/delete/get/filter).
- Migrate data between different databases (same or different database software), using rules to update schema, data, or a
    combination.
- Version Management of all data put into the system (unless skipped)
- Change Management of Version Management commits (done through a series of Action scripts, so highly configurable)

Intended for use on smaller data sets, such as those found in System and Network Operational Configuration Management systems.

Copyright Geoff Howland, 2014.  MIT License.
"""


import sys
import os
import getopt
import pprint

# SchemaMan modules
import utility
from utility.log import Log
from utility.error import *
from utility.path import *

import action


def Main(args=None):
  if not args:
    args = []

  
  long_options = ['dir=', 'verbose', 'help', 'yes']
  
  try:
    (options, args) = getopt.getopt(args, '?hvyd:', long_options)
  except getopt.GetoptError, e:
    Usage(e)
  
  # Dictionary of command options, with defaults
  command_options = {}
  command_options['verbose'] = False
  command_options['always_yes'] = False
  
  
  # Process out CLI options
  for (option, value) in options:
    # Help
    if option in ('-h', '-?', '--help'):
      Usage()
    
    # Verbose output information
    elif option in ('-v', '--verbose'):
      command_options['verbose'] = True
    
    # Always answer Yes to prompts
    elif option in ('-y', '--yes'):
      command_options['always_yes'] = True
    
    # Invalid option
    else:
      Usage('Unknown option: %s' % option)


  # Store the command options for our logging
  utility.log.RUN_OPTIONS = command_options
  
  
  # Ensure we at least have one spec file
  if len(args) < 1:
    Usage('No action specified')
  

  #try:
  if 1:
    action.process.ProcessAction(args[0], args[1:], command_options)
  
  
  #NOTE(g): Catch all exceptions, and return in properly formatted output
  #TODO(g): Implement stack trace in Exception handling so we dont lose where this
  #   exception came from, and can then wrap all runs and still get useful
  #   debugging information
  #except Exception, e:
  else:
    Error({'exception':str(e)}, command_options)


if __name__ == '__main__':
  #NOTE(g): Fixing the path here.  If you're calling this as a module, you have to 
  #   fix the utility/handlers module import problem yourself.
  sys.path.append(os.path.dirname(sys.argv[0]))
  
  Main(sys.argv[1:])



"""
Actions: Test: Test VMCM (Version and Change Managament)

Import a schema into a set of database tables:  database, table, field
"""


# SchemaMan libraries
import schemaman.utility as utility
from schemaman.utility.log import Log
from schemaman.utility.interactive_input import *
import schemaman.datasource as datasource


# This action's command on the CLI and also in the connection_data.actions dict as a key for our data
ACTION = 'test__test_vmcm'


def Action(connection_data, action_input_args):
  """Perform action: Test VMCM (Version and Change Managament)"""
  print 'Test VMCM (Version and Change Managament)'
  
  # Create working Request object
  request = datasource.Request(connection_data, 'ghowland', 'testauth')
  
  # Determine table to operate on
  table = 'owner_group'
  
  record_id = 1
  
  # Get what versions are already available
  versions_available = datasource.RecordVersionsAvailable(request, table, record_id)
  
  # Get original data
  record = datasource.Get(request, table, record_id)
  
  # Make a change
  record['name'] = '%s!' % record['name']
  
  # Save the change un-commited (as a version)
  Log('Set initial changed record in version_working\n\n')
  datasource.SetVersion(request, table, record)
  
  # Get data again (with VM changes applied)
  record_again = datasource.Get(request, table, record_id)
  
  ReadLine('1: Pause for initial set version, type enter to continue: ')
  
  # Get HEAD data (without VM changed applied)
  record_head = datasource.Get(request, table, record_id, use_working_version=False)
  
  # Abandon working versions of data
  Log('Abandon Working version\n\n')
  was_abandoned = datasource.AbandonWorkingVersion(request, table, record_id)
  
  ReadLine('2: Pause for initial set version is abandonded, type enter to continue: ')
  
  # Get data again (with VM changed applied, but no change)
  record_again_again = datasource.Get(request, table, record_id)
  
  # We already made this change once, but if you forgot, this was it:
  # record['name'] = '%s!' % record['name']
  
  # Save the change un-commited (as a version)
  Log('Set second changed record in version_working\n\n')
  datasource.SetVersion(request, table, record)
  
  ReadLine('3: Pause for second set version, type enter to continue: ')
  
  # Get again, see change
  record_again_again_again = datasource.Get(request, table, record_id)
  
  # Commit change
  Log('Commit working version\n\n')
  datasource.CommitWorkingVersion(request)
  
  # Get HEAD data, see change
  record_head_again = datasource.Get(request, table, record_id, use_working_version=False)
  
  # List versions and see where our new version made that change
  versions_available_again = datasource.RecordVersionsAvailable(request, table, record_id)
  



"""
Cache System for SchemaMan

Thread safe.  Locks around all gets and sets of cache data, and all creating of new cache items.

To keep this simple, there is no purge mechanism.  This keeps this code very simple and doesn't
require any maintenance time for cleanup, or extra function calls.

That means dont put long-tail things in this cache.  This is for core items that we may TTL out, but
we dont need to purge, because there arent millions of them.  Thousands of items in memory sitting
around is no problem.
"""


import threading
import time


# Create a pool for caching, and a lock for changing the pool (dicts are thread unsafe)
CACHE_POOL = {}
CACHE_POOL_LOCK = threading.Lock()

# Default Time-To-Live for our cache objects
DEFAULT_TTL = 60 * 5


class NoCacheResultFound:
  """This is used to store a 'no value' result, to differentiate from None, without throwing exceptions"""


class Cache:
  """Object for caching data for a given type."""
  
  def __init__(self, pool_key):
    self.pool_key = pool_key
    
    # We need to lock to get/set data safely
    self.lock = threading.Lock()
    
    # Our cache data, and timeout (time+TTL) for each cache item
    self.data = {}
    self.data_timeout = {}
  
  
  def Get(self, item_key, default_value=NoCacheResultFound):
    """Returns cached data, or default_value"""
    result = default_value
    
    # If our current time is less than the timeout time (time+TTL), if this item exists, get the item
    if time.time() < self.data_timeout.get(item_key, 0):
      result =  self.data.get(item_key, default_value)
    
    return result
  
  
  def Set(self, item_key, value, ttl=DEFAULT_TTL):
    """Sets cached data"""
    aquired_lock = False
    try:
      # Lock the cache and track it
      self.lock.acquire()
      aquired_lock = True
      
      # Set the cache data
      self.data[item_key] = value
      self.data_timeout[item_key] = time.time() + ttl
    
    # Ensure we always release our locks, no matter what
    finally:
      if aquired_lock:
        self.lock.release()


def GetCachePool(pool_key):
  """Returns a Cache class object, from our Cache pool.  Creates it if it doesnt exist."""
  global CACHE_POOL
  global CACHE_POOL_LOCK
  
  
  aquired_pool_lock = False
  try:
    # If we dont have this cache pool object, create it
    if pool_key not in CACHE_POOL:
      # Lock the pool and track it
      CACHE_POOL_LOCK.acquire()
      aquired_pool_lock = True
      
      # Create a Cache object, and put it in our pool key spot
      CACHE_POOL[pool_key] = Cache(pool_key)
  
  # Ensure we always release our locks, no matter what
  finally:
    if aquired_pool_lock:
      CACHE_POOL_LOCK.release()
  
  
  # Return the Cache object, we just enforced it exists
  return CACHE_POOL[pool_key]


def Get(pool_key, item_key, default_value=NoCacheResultFound):
  """Get a value from a cache pool"""
  cache = GetCachePool(pool_key)
  
  value = cache.Get(item_key, default_value=default_value)
  
  return value


def Set(pool_key, item_key, value, ttl=DEFAULT_TTL):
  """Set a value in cache pool"""
  cache = GetCachePool(pool_key)
  
  cache.Set(item_key, value, ttl=ttl)



"""
Handle all SchemaMan datasource specific functions: MySQL
"""

import pprint

import schemaman.datasource as datasource
import schemaman.utility as utility
from schemaman.utility.log import Log
import schemaman.utility.data_control as data_control

import schemaman.datasource.cache as cache

from query import *


# Debugging information logged?
SQL_DEBUG = True


class InvalidArguments(Exception):
  """Something wasnt right with the args."""


class RecordNotFound(Exception):
  """Couldnt find the record."""


def ReleaseConnections(request):
  """Release any connections tied with this request_number"""
  #TODO(g): Flatten this call path
  MySQLReleaseConnections(request)


def TestConnection(request):
  """Create a schema, based on a spec"""
  Log('MySQL: Test Connection: %s: %s' % (request.connection_data['alias'], request.request_number))
  
  connection = GetConnection(request)
  
  #TODO(g): Make a better connection test?  SELECT 1?
  result = connection.Query('SHOW TABLES')
  
  return result


def CreateSchema(request):
  """Create a schema, based on a spec"""
  raise Exception('TBD...')
  pass


def ExtractSchema(request):
  """Export a schema, based on a spec, or everything"""
  Log('MySQL: Extract Schema: %s: %s' % (request.connection_data['alias'], request.request_number))
  
  connection = GetConnection(request, server_id=request.server_id)
  
  # Pack all our schema data in here
  data = {}
  
  tables = connection.Query("SHOW TABLES")
  
  for table in tables:
    key = table.keys()[0]
    table_name = table[key]
    
    data[table_name] = {}
    
    fields = connection.Query("DESC `%s`" % table_name)
    
    for item in fields:
      # schema_table_field: {u'Extra': u'auto_increment', u'Default': None, u'Field': u'id', u'Key': u'PRI', u'Null': u'NO', u'Type': u'int(11)'}
      print '%s: %s' % (table_name, item)
      
      data[table_name][item['Field']] = {}
      field = data[table_name][item['Field']]
      
      field['name'] = item['Field']
      field['default'] = item['Default']
      
      # If this is not a sized type
      if '(' not in item['Type']:
        field['type'] = item['Type']
        field['size1'] = None
        field['size2'] = None
      
      # Else, get the sizes for this type, as well as the type name
      else:
        (type_name, type_args) = item['Type'].split('(', 1)
        type_args = type_args.replace(')', '').split(',')
        
        field['type'] = type_name
        
        # Type Size
        if type_args:
          field['size1'] = int(type_args[0])
          
          if len(type_args) >= 2:
            field['size2'] = int(type_args[1])
          else:
            field['size2'] = None
      
      # Allow Null?
      if item['Null'] == 'YES':
        field['allow_null'] = True
      else:
        field['allow_null'] = False
      
      # Primary Key?
      if item['Key'] == 'PRI':
        field['pkey'] = True
      else:
        field['pkey'] = False
        
      # Multiple Primary Key?
      if item['Key'] == 'MUL':
        #TODO(g): Add multi-key support.
        print 'TODO(g): Add multi-key support.'
    
      # Get the unique constraints too.  These are what allow us to constrain our data for uniqueness, so it's important.  Separate from PKEY indexs.
  
  return data


def ExportSchema(request):
  """Export a schema, based on a spec, or everything"""
  raise Exception('TBD...')
  pass


def UpdateSchema(request):
  """Update a schema, based on a spec.
  
  Can go 'forward' or 'backwards' for version control, its still updating.
  """
  raise Exception('TBD...')
  pass


def ExportData(request):
  """Export/dump data from this datasource, based on spec, or everything"""
  raise Exception('TBD...')
  pass


def ImportData(request, drop_first=False, transaction=False):
  """Import/load data to this datasource, based on spec, or everything.
  
  Args:
    drop_first: boolean, optional: If true, all data is dropped/deleted before
        the import occurs, otherwise it is an update.  Defaults to false to
        preserve data.
    transaction: boolean, optional: If true, import is done as a single
        transaction.  Defaults to False to avoid extra memory and slowness.
  
  Returns: None
  """
  raise Exception('TBD...')
  pass


def GetUser(request, username=None, use_cache=True):
  """Returns user.id (int)"""
  # We allow an explicit username, but otherwise use the requester's username
  if not username:
    username = request.username
  
  # If we want to use the cache, and we have this is cache, return it
  if use_cache:
    user = cache.Get('user_by_name', username)
    if user != cache.NoCacheResultFound:
      return user
  
  # Get a connection
  connection = GetConnection(request)
  
  #TODO(g): Need to specify the schema (DB) too, otherwise this is wrong...  Get from the request datasource info?  We populated, so we should know how it works...
  sql = "SELECT * FROM `user` WHERE name = %s"
  result = connection.Query(sql, [username])
  if not result:
    raise Exception('Unknown user: %s' % username)
  
  user = result[0]

  # Save this is cache
  cache.Set('user_by_name', username, user)
  cache.Set('user_by_id', user['id'], user)

  return user


def GetUserById(request, user_id, use_cache=True):
  """Returns user record (dict)"""
  # Get a connection
  connection = GetConnection(request)
  
  # If we want to use the cache, and we have this is cache, return it
  if use_cache:
    user = cache.Get('user_by_id', user_id)
    if user != cache.NoCacheResultFound:
      return user
  
  #TODO(g): Need to specify the schema (DB) too, otherwise this is wrong...  Get from the request datasource info?  We populated, so we should know how it works...
  sql = "SELECT * FROM `user` WHERE id = %s"
  result = connection.Query(sql, [user_id])
  if not result:
    raise Exception('Unknown user: %s' % user_id)
  
  user = result[0]

  # Save this is cache
  cache.Set('user_by_name', user['name'], user)
  cache.Set('user_by_id', user['id'], user)
  
  return user


def GetInfoSchema(request):
  """Returns the record for this schema data (schema)"""
  # Get this cached result, and return it if available
  cache_result = cache.Get('schema', request.connection_data['datasource']['database'])
  if cache_result != cache.NoCacheResultFound:
    return cache_result

  # Get a connection
  connection = GetConnection(request)
  
  # Get the schema name from our request.datasource.database
  database_name = request.connection_data['datasource']['database']
  
  sql = "SELECT * FROM `schema` WHERE `name` = %s"
  result_schema = connection.Query(sql, [database_name])
  if not result_schema:
    raise Exception('Unknown schema: %s' % database_name)
  
  schema = result_schema[0]
  
  # Set this cache result
  cache.Set('schema', request.connection_data['datasource']['database'], schema)
  
  return schema


def GetInfoSchemaTable(request, schema, table, filter_key='name'):
  """Returns the record for this schema table data (schema_table)"""
  # Get this cached result, and return it if available
  cache_result = cache.Get('schema_table__%s' % filter_key, (schema['id'], table))
  if cache_result != cache.NoCacheResultFound:
    return cache_result

  # Get a connection
  connection = GetConnection(request)
  
  #TODO(g): Need to specify the schema (DB) too, otherwise this is wrong...  Get from the request datasource info?  We populated, so we should know how it works...
  sql = "SELECT * FROM `schema_table` WHERE schema_id = %%s AND %s = %%s" % filter_key
  result_schema_table = connection.Query(sql, [schema['id'], table])
  if not result_schema_table:
    raise Exception('Unknown schema_table: %s: %s' % (request.connection_data['datasource']['database'], table))
  
  schema_table = result_schema_table[0]
  
  # Save the cache result
  cache.Set('schema_table__%s' % filter_key, (schema['id'], table), schema_table)
  
  return schema_table


def GetInfoSchemaAndTable(request, table_name):
  """Returns the record for this schema table data (schema_table)
  
  This is a helper function, calls GetInfoSchema() and GetInfoSchemaTable()
  """
  # Get this cached result, and return it if available
  cache_result = cache.Get('schema_and_table', (request.connection_data['datasource']['database'], table_name))
  if cache_result != cache.NoCacheResultFound:
    return cache_result

  schema = GetInfoSchema(request)
  
  schema_table = GetInfoSchemaTable(request, schema, table_name)
  
  # Save the cache result
  cache.Get('schema_and_table', (request.connection_data['datasource']['database'], table_name), (schema, schema_table))
  
  return (schema, schema_table)


def GetInfoSchemaAndTableById(request, schema_table_id):
  """Returns the record for this schema table data (schema_table)
  
  This is a helper function, calls GetInfoSchema() and GetInfoSchemaTable()
  """
  schema = GetInfoSchema(request)
  
  schema_table = GetInfoSchemaTable(request, schema, schema_table_id, filter_key='id')
  
  return (schema, schema_table)


def GetInfoSchemaTableField(request, schema_table, name):
  """Returns the record for this schema field data (schema_table_field)"""
  # Get a connection
  connection = GetConnection(request)
  
  sql = "SELECT * FROM `schema_table_field` WHERE schema_table_id = %s AND name = %s"
  result_schema_table_field = connection.Query(sql, [schema['id'], table])
  if not result_schema_table_field:
    raise Exception('Unknown schema_table_field: %s: %s: %s' % (request.connection_data['datasource']['database'], table, name))
  
  schema_table = result_schema_table_field[0]
  
  return schema_table


def RecordVersionsAvailable(request, table, record_id, user=None):
  """List all of the historical and currently available versions available for this record.
  
  Looks at 3 tables to figure this out: version_pending_log (un-commited changes),
      version_commit_log (commited changes), version_working (single user changes)
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, primary key (ex: `id`) of the record in this table.  Use Filter() to use other field values
    username: string (default None), if not None, this is a specific user to check versions.  Otherwise the
        request.username is used.
    
  Returns: list of dicts, dicts have 'id' and 'name' fields.
  """
  # Get a connection
  connection = GetConnection(request)
  
  # Get the schema name from our request.datasource.database
  database_name = request.connection_data['datasource']['database']
  
  # Get the schema and table info
  (schema, schema_table) = GetInfoSchemaAndTable(request, table)  
  
  # version_pending_log
  sql = "SELECT * FROM `version_pending_log` WHERE schema_id = %s AND schema_table_id = %s AND record_id = %s ORDER BY id"
  result_changelist = connection.Query(sql, [schema['id'], schema_table['id'], record_id])
  
  # version_commit_log
  sql = "SELECT * FROM `version_commit_log` WHERE schema_id = %s AND schema_table_id = %s AND record_id = %s ORDER BY id"
  result_commit = connection.Query(sql, [schema['id'], schema_table['id'], record_id])
  
  # version_working
  sql = "SELECT * FROM `version_working` WHERE user_id = %s"
  result_working = connection.Query(sql, [user['id']])
  
  
  # Compile the final result list, from the found results
  result = []
  
  # Commited versions
  for item in result_commit:
    data = {'id': item['id'], 'name':'Commit Number: %s' % item['id']}
    result.append(data)
  
  # Change List versions
  for item in result_changelist:
    data = {'id': item['id'], 'name':'Pending Change Number: %s' % item['id']}
    result.append(data)
  
  # Working set
  if result_working:
    working = result_working[0]
    
    data = utility.path.LoadYamlFromString(working['data_yaml'])
    
    if schema['id'] in data:
      if schema_table['id'] in data[schema['id']]:
        if record_id in data[schema['id']][schema_table['id']]:
          item = {'id':'working', 'name':'Working Version: %s' % user['name']}
          result.append(item)
  
  return result


def CommitWorkingVersion(request):
  """Immediately takes everything in this user's Working Set and creates a change list.
  
  This is the short-cut for Change Management, so we don't need all the steps and still have versionining.
  
  See also: CreateChangeList() and CreateChangeListFromWorkingSet() and CommitWorkingVersionSingleRecord()
  """
  working_version = GetUserVersionWorkingRecord(request)
  
  # Make a single record entry in the version_pending table, do all the work as we normally would (increments the PKEY, etc)
  version_number = CreateChangeList(request, working_version)
  
  # "commit" the changelist into version_commit, which will also put the data into the direct DB tables
  result = CommitChangeList(request, version_number)
  
  # Clean up the working version
  #TODO(g): Add commit=False to all of these and then Commit() after all of this?  Yes, do it when coming back to this.
  #   We need a CloneRequest() type method to get a new request connection object, so we are commiting the version data separately from the actual data.
  Delete(request, 'version_working', working_version['id'])
  
  return result


def CommitWorkingVersionSingleRecord(request, table, record_id):
  """TODO: Commit only a single record out of the working version.  Not all the working version records...
  
  Immediately takes a working version record and commits it, moving it through the rest of change management.
  
  This is the short-cut for Change Management, so we don't need all the steps and still have versionining.
  
  See also: CreateChangeList() and CreateChangeListFromWorkingSet()
  """
  raise Exception('TBD...')

  working_version = GetUserVersionWorkingRecord(request)
  
  record = GetRecordFromVersionRecord(request, working_version, table, record_id)

  #TODO(g): Format the record so it is wrapped in it's schema/schema_table/record dicts, as it is a full change now
  pass
  wrapped_record = None
  pass
  
  # Make a single record entry in the version_pending table, do all the work as we normally would (increments the PKEY, etc)
  version_number = CreateChangeList(request, wrapped_record)
  
  # "commit" the changelist into version_commit, which will also put the data into the direct DB tables
  result = CommitChangeList(request, version_number)
  
  # Clean up the working version
  raise Exception('Clean up the working version, we just left that there after making the Change List and commiting it...  Also, transactions?  Make sure thats OK too.  Commit on end?  New request for versions?  Clone request to get new connection/transaction?  Yes.')
  
  return result


def CommitChangeList(request, version_number):
  """Commit a pending change list (version_pending) to the final data, and put in version_commit.
  
  Also updates version_pending_log (removes entries), and version_commit_log (adds entries).
  
  This function works as a single datasbase transaction, so it cannot leave the data in an inconsistent state.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    version_number: int, this is the version number in the version_pending.id
  
  Returns: None
  """
  # Get the specified change list record
  record = Get(request, 'version_pending', version_number, use_working_version=False)
  
  # Create the new commit record to be inserted
  data = {'user_id':request.user['id'], 'data_yaml':record['data_yaml']}
  
  # Insert into version_commit
  version_commit_id = SetDirect(request, 'version_commit', data, commit=False)
  
  # Create the commit_log data
  CreateVersionLogRecords(request, 'version_commit', version_commit_id, data, commit=False)
  
  # Remove the version_pending_log row
  DeleteFilter(request, 'version_pending_log', {'version_pending_id':version_number}, commit=False)
  
  # Remove the version_pending row
  Delete(request, 'version_pending', record['id'], commit=False)
  
  # Make the change to the tables that are effected
  CommitVersionRecordToDatasource(request, version_commit_id, record, commit=False)
  
  # Commit the request
  Commit(request)


def CommitVersionRecordToDatasource(request, version_commit_id, change_record, commit=True):
  """Commit a change from the version_commit table into the real (non-versioning) datasource tables.
  
  This should not be called from outside this library, mostly because there is no reason to and it
  requires correct internal state to keep things in order.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    version_number: int, this is the version number in the version_pending.id
    change_record: dict, `version_commit` table row data
    commit: boolean (default True), if True any queries that could be commited will be (single query transaction), if False then a later Commit() will be required
  
  Returns: None
  """
  # Parse the YAML for our cahnge data
  update_data = utility.path.LoadYamlFromString(change_record['data_yaml'], {})
  delete_data = utility.path.LoadYamlFromString(change_record['delete_data_yaml'], {})
  admin_data = utility.path.LoadYamlFromString(change_record['admin_data_yaml'], {})
  rollback_data = utility.path.LoadYamlFromString(change_record['rollback_data_yaml'], {})
  
  
  # Get all our items as a flat dict of our changes, as it's easier to iterate over
  update_items = data_control.NestedDictsToSingleDict(update_data, 3)
  
  
  # Track dependencies here
  dependencies = {}
  dependency_results = {}
  dependency_update = {}
  
  # We start out with deferred items being everything, then we remove them, as possible, and loop until they are clear
  deferred_items = dict(update_items)
  
  # Create sorted record_keys list, we know these as they clear their dependencies, so automatically sorted
  sorted_items = []
  
  # We will keep looping until we either have no deferred items, or we couldnt find any more dependencies (circular references)
  change_occurred = True
  while deferred_items and change_occurred:
    # This must be set to True each time, or we will not loop again, and if there are still deferred items, we will error
    change_occurred = False
    
    # Loop over our update_items and create a dependency structure
    deferred_keys = deferred_items.keys()
    for record_key in deferred_keys:
      (schema_id, schema_table_id, record_id) = record_key
      (schema, schema_table) = GetInfoSchemaAndTableById(request, schema_table_id)
      record_data = deferred_items[record_key]
      
      # If we are a new record (id<0), add ourselves to the dependencies dict
      if record_data.get('id', 0) < 0:
        if schema_table['name'] not in dependencies:
          dependencies[schema_table['name']] = {}
        
        # Create our entry in the dependencies, so we can be easily referenced against fields of dependent records
        dependencies[schema_table['name']][record_data['id']] = record_key
      
      
      # Test the fields for dependencies first
      field_dependencies = {}
      
      # Check if this has any field dependencies
      for (field, field_value) in record_data.items():
        if field.endswith('_id') and type(field_value) == int and field_value < 0:
          # Get the table name (potentially), by stripping off the '_id' characters
          table_name = field[:-3]
          
          # Ensure this is actually a table name we know about
          try:
            # If this doesnt throw an exception, this is a real table name, and so this is a join table, and we should add it as a dependency
            GetInfoSchemaAndTable(request, table_name)
          except Exception, e:
            # This is not a table name, so we dont need to track it as a dependency, skip this field
            continue
          
          # Check to see if we already have the dependency we need listed
          if table_name in dependencies and field_value in dependencies[table_name]:
            # If we already have it, do nothing.  We will see this doesnt have any dependent fields (if it doesnt), and remove it from the deferred queue after this block
            if record_key not in dependency_update:
              dependency_update[record_key] = {}
            
            # Save what we want to collect once we are ready to be inserted, later (after our required records are inserted)
            dependency_update[record_key][field] = {'table': table_name, 'value': field_value}
            # print 'Recording lookup for later: %s: %s: %s: %s' % (record_key, field, table_name, field_value)
          
          # Else, this dependency hasnt been found yet, so keep looking for more dependent fields
          else:
            # This is a dependent field
            field_dependencies[table_name] = field_value
            # print 'Found dependent field: %s: %s: %s: %s' % (record_key, field, table_name, field_value)
      
      
      # If we dont have any field dependencies, then we can be removed from the deferred items, as we are set
      if not field_dependencies:
        # print 'No dependencies, so removing: %s' % str(record_key)
        del deferred_items[record_key]
        
        # We modified the deferred items, so change has occurred
        change_occurred = True
        
        # We can sort these items on the order they clear being deferred, this guarantees we know their dependencies already, so is perfect
        sorted_items.append(record_key)
  
  
  # print '\n\n::: Dependencies:\n%s\n\n' % pprint.pformat(dependencies)
  # print '\n\n::: Dependency Updates:\n%s\n\n' % pprint.pformat(dependency_update)
  # print '\n\n::: Sorted Items:\n%s\n\n' % pprint.pformat(sorted_items)
  
  # If we had a circular reference, we still have deferred items, but we couldnt change anything in a pass, so we are stuck
  if not change_occurred and deferred_items:
    raise Exception('Circular reference found in deferred items, could not create dependency list: %s' % deferred_items)

  
  # Process all our sorted items, which takes care of getting dependencies when needed
  for record_key in sorted_items:
    (schema_id, schema_table_id, record_id) = record_key
    (schema, schema_table) = GetInfoSchemaAndTableById(request, schema_table_id)
    record = update_items[record_key]

    # If this is a New Record (id<0), remove the 'id' field, as we will auto_increment it into existance and uet the updated_record_id from that
    if record.get('id', 0) < 0:
      del record['id']
    
    # Rollback: Get the real record, this is None if it doesnt exist, which is also what we want.  It works for the positive and the negative existance cases.
    real_record = Get(request, schema_table['name'], record_id)
    data_control.EnsureNestedDictsExist(rollback_data, [schema_id, schema_table_id, record_id], real_record)
    
    # If this record requires any updates
    if record_key in dependency_update:
      for (field, field_data) in dependency_update[record_key].items():
        # Get our dependency record_key, we have to look it up
        dependency_record_key = dependencies[field_data['table']][field_data['value']]
        
        # Get our updated value
        record[field] = dependency_results[dependency_record_key]
    
    # If we got a real record, then overly our changes onto it, so we can save them properly
    if real_record:
      new_record = dict(real_record)
      new_record.update(record)
      record = new_record
    
    # Directly save this into table it was intended to be in.  Update it's own ID, so we convert New Records into existing ones
    # print '\nTODO: Set Direct: %s: %s\n' % (schema_table['name'], record)
    record['id'] = SetDirect(request, schema_table['name'], record, commit=commit)
    
    # Save all the updated record IDs here, if we need them, we will come and get them
    dependency_results[record_key] = record['id']
    
    # Put this record back into the update_data section, so that we keep the updated IDs, if they have been
    update_data[schema_id][schema_table_id][record_id] = record
  
  
  # Delete the specified records as well
  for (schema_id, schema_data) in delete_data.items():
    for (schema_table_id, schema_table_data) in schema_data.items():
      (schema, schema_table) = GetInfoSchemaAndTableById(request, schema_table_id)
      
      for record_id in schema_table_data:
        # Get the existing record, so we can store it for rollback
        real_record = Get(request, schema_table['name'], record_id)
        data_control.EnsureNestedDictsExist(rollback_data, [schema_id, schema_table_id, record_id], real_record)
        
        # Delete the Real record
        Delete(request, schema_table['name'], record_id)

  print '\n\n::: Roll Back data:\n%s\n\n' % pprint.pformat(rollback_data)
  
  # Save the updated version_commit results back into the DB, with our rollback data and any updated fields from dependencies
  change_record['data_yaml'] = utility.path.DumpYamlAsString(update_data)
  change_record['rollback_data_yaml'] = utility.path.DumpYamlAsString(rollback_data)
  SetDirect(request, 'version_commit', change_record)


def CreateVersionLogRecords(request, version_table, version_id, data, commit=True):
  """Create all the rows needed in the `version_*_log` tables (specified by table) for the data
  
  Version table information to stored in data['data_yaml'] as JSON encoded dict, which is keyed
  on the schema.id (int), then a dict keyed on the schema_table.id (int), then the field names (string)
  for the final dict, with values of the table field values (varying types).
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    version_table: string, name of table to operate on
    version_id: int, the PKEY of the table record we are creating log records for
    data: dict, record that we want to store in the table row
    commit: boolean (default True), if True any queries that could be commited will be (single query transaction), if False then a later Commit() will be required
  
  Returns: list of ints, all of the PKEYs for the log rows we inserted
  """
  # We will return a list of ints, which are all the row `id` field values (PKEYs) for the table records
  log_row_ids = []
  
  
  Log('Change Log: %s' % data, logging.DEBUG)
  
  change = utility.path.LoadYamlFromString(data['data_yaml'])
  
  Log('Writing version log records for: %s' % change, logging.DEBUG)
  
  # Process all the schema table fields we need version logs for
  for (schema_id, schema_tables) in change.items():
    
    Log('Change: %s : %s' % (schema_id, schema_tables))
    
    for (schema_table_id, records) in schema_tables.items():
      for record_id in records:
        # Create our reference field, based on the table name (ex: version_commit_id)
        reference_field = '%s_id' % version_table
        version_table_log = '%s_log' % version_table
        
        # Create the log record data to insert
        log_data = {reference_field: version_id, 'schema_id':schema_id, 'schema_table_id':schema_table_id, 'record_id':record_id}
        
        # Directly save this into the `version_*_log` table, with the commit flag specified
        SetDirect(request, version_table_log, log_data, commit=commit)
  
  return log_row_ids


def CreateChangeList(request, data):
  """Create a change list from the given table and record_id from the Working Set.
  
  This ensures we always have at least something in a change list, so we dont end up with empty ones where we dont know what
  went wrong.  We will always have some indication of what was going on if we find a change list.
  
  See also: CommitWorkingVersion() and CreateChangeListFromWorkingSet()
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, record that we want to store in the table row
  
  Returns: int, version_number for this change list
  """
  Log('Create change list: %s' % data)
  
  # data_yaml = utility.path.DumpYamlAsString(data)
  
  # Create the changelist record
  record = {'user_id':request.user['id'], 'data_yaml':data['data_yaml']}
  
  # Create this version pending change, and get the version number
  version_number = SetDirect(request, 'version_pending', record)
  
  # Create the version log for this pending change
  CreateVersionLogRecords(request, 'version_pending', version_number, record)
  
  return version_number


def CreateChangeListFromWorkingSet(request):
  """Create a Change List from the entire Working Set of version records.
  
  This is a fast way to prepare to commit everything that is currently being worked on.
  
  See also: CommitWorkingVersion() and CreateChangeList()
  
  Returns: int, version_number for this change list
  """
  raise Exception('TBD...')
  
  return version_number


def AbandonWorkingVersion(request, table, record_id):
  """Abandon any current edits in version_working for this record.
  
  This does not effect Change Lists that are created, which must either be editted (removing record), or else
  the entire change list must be abandonded.

  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, record `id` field, primary key
  
  Returns: boolean, True if there was a working version to abandon, False if there was no working version to abandon
  """
  # Get this user's working version record
  try:
    record = GetUserVersionWorkingRecord(request)
  
  except datasource.VersionNotFound, e:
    return False
  
  
  # Extract the data_yaml payload
  change = utility.path.LoadYamlFromString(record['data_yaml'])

  # Format record key
  #TODO(g): Do this properly with the above dynamic PKEY info.  Is this good enough because we take record_id?  Maybe this needs to already be turned into the data_key?  This definitely needs to be a First Class Citizen in schemaman
  data_key = record_id
  
  # Get a connection
  connection = GetConnection(request)
  
  # Get the schema and table info
  (schema, schema_table) = GetInfoSchemaAndTable(request, table)


  # Add this set data to the version change record, if it doesnt exist
  if schema['id'] not in change:
    raise Exception('This user does not have the specified record in their working version: %s: %s: %s: %s: %s' % (request.username, table, schema['id'], schema_table['id'], record_id))
  
  # Add this table to the change record, if it doesnt exist
  if schema_table['id'] not in change[schema['id']]:
    raise Exception('This user does not have the specified record in their working version: %s: %s: %s: %s: %s' % (request.username, table, schema['id'], schema_table['id'], record_id))
  
  # Delete the record from the working version data
  del change[schema['id']][schema_table['id']]
  
  # Put this change record back into the version_change table, so it's saved
  record['data_yaml'] = utility.path.DumpYamlAsString(change)

  
  # Save the change record
  result_record = SetDirect(request, 'version_working', record)
  
  return True


def AbandonChangeList(request, change_list_id):
  """Abandon an open Change List, change_list_id==version_change.id
  """
  raise Exception('TBD...')
  
  return result


def GetUserVersionWorkingRecord(request, user_id=None):
  """Returns a dict (row) from the version_working table for this request.user"""
  # If we werent given an explicit user, use the request user
  if not user_id:
    user_id = request.user['id']
  
  # Get a connection
  connection = GetConnection(request)
  
  # Get the schema and table info
  (schema, schema_table) = GetInfoSchemaAndTable(request, 'version_working')  
  
  # Get the current working record for this user (if  any)
  sql = "SELECT * FROM version_working WHERE user_id = %s"
  result = connection.Query(sql, [user_id])
  
  # If we have a current change version record, use that
  if result:
    record = result[0]
  
  # Else, we dont have one yet, so create one
  else:
    raise datasource.VersionNotFound('No version working data exists for user: %s' % request.username)
  
  return record


def GetRecordFromVersionRecord(request, version_record, table, record_id):
  """Get a record from a version record (could be version_working, changelist or commit row record).
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, record `id` field, primary key
    
  Returns: dict, row record from the version record (stored in data_yaml field)
  """
  # Get the JSON payload from the version record
  change = utility.path.LoadYamlFromString(version_record['data_yaml'])
  
  Log('GetRecordFromVersionRecord: Change: %s' % change, logging.DEBUG)
  
  # Format record key
  #TODO(g): Do this properly with the above dynamic PKEY info
  data_key = record_id
  
  # Get the schema and table info
  (schema, schema_table) = GetInfoSchemaAndTable(request, table)
  
  # Add this set data to the version change record, if it doesnt exist
  if schema['id'] not in change:
    raise datasource.RecordNotFound('Could not find version record: %s: %s: %s: %s: %s' % (request.username, table, schema['id'], schema_table['id'], record_id))
  
  # Add this table to the change record, if it doesnt exist
  if schema_table['id'] not in change[schema['id']]:
    raise datasource.RecordNotFound('Could not find version record: %s: %s: %s: %s' % (request.username, table, schema['id'], schema_table['id'], record_id))
  
  # Ensure the record is in the table
  if data_key not in change[schema['id']][schema_table['id']]:
    raise datasource.RecordNotFound('Could not find version record: %s: %s: %s: %s: %s' % (request.username, table, schema['id'], schema_table['id'], record_id))
  
  return change[schema['id']][schema_table['id']][data_key]
  

def Commit(request):
  """Commit a datasource transaction that is in the middle of a transaction.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
  
  Returns: None
  """
  connection = GetConnection(request)
  
  connection.Commit()


def AbandonCommit(request):
  """Abandon Commit a datasource transaction that is in the middle of a transaction.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
  
  Returns: None
  """
  connection = GetConnection(request)
  
  connection.AbandonCommit()


def Set(request, table, data, version_management=True, commit_version=False, version_number=None, noop=False, update_returns_id=True, debug=SQL_DEBUG, commit=True):
  """Put (insert/update) data into this datasource.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, record that we want to store in the table row
    commit_version: boolean (default False), if True, this will automatically commit this version as part of this set (in version_commit)
    version_number: int (default None), if an int, this is the version number in the version_change or version_commit
        tables.  version_change is scanned before version_commit, as these are more likely to be requested.
    noop: boolean (default False), if True do not actually query the database, (no operation)
    update_returns_id: boolean (default True), if True UPDATE will return the PKEY (ex: `id`) keeping the same results that INSERT does
    debug: boolean, if True will log more verbosely
    commit: boolean (default True), if True any queries that could be commited will be (single query transaction), if False then a later Commit() will be required

  Return: int or None, If commit_version==True then this is the real table's PKEY int, else None
  """
  # If we are directly working with database tales
  if not version_management:
    SetDirect(request, table, data, noop=noop, update_returns_id=update_returns_id, debug=debug, commit=commit)
  
  # Else, this work should be in the version_* tables
  else:
    SetVersion(request, table, data, commit_version=commit_version, version_number=version_number, noop=noop, update_returns_id=update_returns_id, debug=debug)


def SetVersion(request, table, data, commit_version=False, version_number=None, noop=False, update_returns_id=True, debug=SQL_DEBUG):
  """Put (insert/update) data into this datasource.  Writes into version management tables (working or changelist if version_number is specified)
  
  Works as a single transaction, as version data is always commited into the version_* tables.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, record that we want to store in the table row
    commit_version: boolean (default False), if True, this will automatically commit this version as part of this set (in version_commit)
    version_number: int (default None), if an int, this is the version number in the version_change or version_commit
        tables.  version_change is scanned before version_commit, as these are more likely to be requested.
    update_returns_id: boolean (default True), if True UPDATE will return the PKEY (ex: `id`) keeping the same results that INSERT does
    debug: boolean, if True will log more verbosely

  Return: int or None, If commit_version==True then this is the real table's PKEY int, else None
  """
  # Get a connection
  connection = GetConnection(request)
  
  # Get the schema and table info
  (schema, schema_table) = GetInfoSchemaAndTable(request, table)

  print '\n\n** SetVersion: %s: %s\n\n' % (table, data)

  
  # If this is a working version (no version number)
  if not version_number:
    #TODO(g): Use functions like GetUserVersionWorkingRecord()?  Maybe not, I dont think it will be helpful.  Wait until later to determine how to unify this better, once I've written all the functions...  Still early...
    # Get the current working record for this user (if  any)
    sql = "SELECT * FROM version_working WHERE user_id = %s"
    result = connection.Query(sql, [request.user['id']])
    
    # Which table are we working with?  No version is working table
    version_table = 'version_working'

  
  # Else, there is a version_number, so work with the version_pending record
  else:
    # Get the current working record for this user (if  any)
    #NOTE(g):SECURITY: Im not forcing that only the owner can write these here.  Is that wrong?  Should I?  I think there should be a different authorization phase...
    sql = "SELECT * FROM version_pending WHERE id = %s"
    result = connection.Query(sql, [version_number])
    
    # Fail if we cant find this record
    if not result:
      raise Exception('The pending change was not found: %s' % version_number)
    
    # Which table are we working with?  We have a version number, so it's changelist
    version_table = 'version_pending'
  
  
  
  # Get the record data set up
  if result:
    record = result[0]
    change = utility.path.LoadYamlFromString(record['data_yaml'])
    delete_change = utility.path.LoadYamlFromString(record['delete_data_yaml'])
    if not change:
      change = {}
    if not delete_change:
      delete_change = {}
  
  # Else, we dont have one yet, so create one (this can only execute on working, because we fail if we have not result with pending)
  else:
    record = {'user_id':request.user['id'], 'data_yaml':{}}
    change = {}
    delete_change = {}
  
  
  # Get the primary key information for this table, and determine how to format our key
  pass

  # Format record key
  #TODO(g): Do this properly with the above dynamic PKEY info
  data_key = data['id']
  
  # Add this set data to the version change record, if it doesnt exist
  if schema['id'] not in change:
    change[schema['id']] = {}
  
  # Add this table to the change record, if it doesnt exist
  if schema_table['id'] not in change[schema['id']]:
    change[schema['id']][schema_table['id']] = {}
  
  # Readability variable
  change_table = change[schema['id']][schema_table['id']]
  
  print '\n\nSetting data in table: %s: %s: %s\nChange Table: %s\nData: %s\n\n' % (schema['id'], schema_table['id'], data_key, change_table, data)
  
  # Add this specific record, or update it if it already exists
  if data_key in change_table:
    change_table[data_key].update(data)
  else:
    change_table[data_key] = data
  
  print '\nAfter Setting Data: %s\n\n' % change_table

  
  # If we had an entry in the delete_change list for this record, remove that.  Any position change wipes out a delete, for obvious reasons.
  if schema['id'] in delete_change:
    if schema_table['id'] in delete_change[schema['id']]:
    
      # If we have this record's ID in our delete list, remove it
      if data['id'] in delete_change[schema['id']][schema_table['id']]:
        delete_change[schema['id']][schema_table['id']].remove(data['id'])
 

  # Get the Real record (if it exists), so we only store fields that are different.  If all fields are the same, we store nothing
  real_record = Get(request, table, data['id'], use_working_version=False)
  
  # If we have a Real record, then remove any matching fields
  if real_record:
    for (real_key, real_value) in real_record.items():
      # If our change data has this key
      if real_key in change_table[data_key]:
        # If the key is the same value as the Real record value, then we dont need it versioned, because it hasnt changed, and it isnt the 'id' key
        #   Also, delete if the real value is NULL, and we have an empty string (no change)
        if real_value == change_table[data_key][real_key] or (real_value == None and change_table[data_key][real_key] == ''):
          del change_table[data_key][real_key]


  # Clean up any unused data structures, so we dont have a bunch of junk hanging around
  datasource.CleanEmptyVersionData(change)
  datasource.CleanEmptyVersionData(delete_change)

  
  # Put this change record back into the version_change table, so it's saved
  record['data_yaml'] = utility.path.DumpYamlAsString(change)
  record['delete_data_yaml'] = utility.path.DumpYamlAsString(delete_change)
    
  # Save the change record
  result_record = SetDirect(request, version_table, record)
  
  return result_record


def SetDirect(request, table, data, noop=False, update_returns_id=True, debug=SQL_DEBUG, commit=True):
  """Put (insert/update) data into this datasource.  Directly writes to database.
  
  Works as a single transaction if commit==True.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, record that we want to store in the table row
    noop: boolean (default False), if True do not actually query the database, (no operation)
    update_returns_id: boolean (default True), if True UPDATE will return the PKEY (ex: `id`) keeping the same results that INSERT does
    debug: boolean, if True will log more verbosely
    commit: boolean (default True), if True any queries that could be commited will be (single query transaction), if False then a later Commit() will be required
  """
  # INSERT values into a table, and if they already exist, perform an UPDATE on the fields
  base_sql = "INSERT INTO `%s` (%s) VALUES (%s) ON DUPLICATE KEY UPDATE %s"
  
  # Wrap all keys in backticks, so they cannot conflict with SQL keywords
  keys = data.keys()
  keys.sort()
  keys_ticked = []
  update_sets = []
  values = []
  value_format_list = []
  
  
  # Get our backticked wrapped insert keys, our value list, and our update setting
  for count in range(0, len(keys)):
    # Back tick column names
    ticked_key = '`%s`' % keys[count]
    keys_ticked.append(ticked_key)
    
    # Value list
    values.append(data[keys[count]])
    value_format_list.append('%s')
    
    # Update keys will reference the insert keys, so we dont have to specify the data twice (SQL does it)
    #TODO(g): Should I remove the primary key from this?  Not sure it's necessary.  Remove comment when proven it works without removing it (simpler)...
    update_sets.append('%s=VALUES(%s)' % (ticked_key, ticked_key))
  
  
  # Build out strings to insert into our base_sql
  insert_columns = ', '.join(keys_ticked)
  value_format_str = ', '.join(value_format_list)
  update_sql = ', '.join(update_sets)
  
  
  # Create our final SQL
  sql = base_sql % (table, insert_columns, value_format_str, update_sql)
  
  # Get a connection
  connection = GetConnection(request)
  
  
  # Query.  Will return a row_id (int) for INSERT, and None for Update
  if not noop:
    print sql
    result = connection.Query(sql, values, commit=commit)
    
    # If we did an Update, we really want the 'id' field returned, like INSERT does (consistency and not having to do this all the time after an update)
    if result == 0 and update_returns_id:
      # If we have a primary 'id' key, use that
      if 'id' in data:
        result = Get(request, table, data['id'])
      
      # Else, use all the fields to get it explicitly
      else:
        # This should always return a single dict in a list, due to our uniqueness constraints
        rows = Filter(request, table, data)
        result = rows[0]['id']
    
  else:
    Log('Query not run, noop = True')
    result = None
  
  return result


def Get(request, table, record_id, version_number=None, use_working_version=True):
  """Get (select single record) from this datasource.
  
  Can be a 'view', combining several lower level 'tables'.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, primary key (ex: `id`) of the record in this table.  Use Filter() to use other field values
    version_number: int (default None), if an int, this is the version number in the version_change or version_commit
        tables.  version_change is scanned before version_commit, as these are more likely to be requested.
    use_working_version: boolean (default True), if True and version_number==None this will also look at any
        version_working data and return it instead the head table data, if it exists for this user.
  
  Returns: dict, single record key/values
  """
  # Get a connection
  connection = GetConnection(request)
  
  found_version_record = None
  
  # If we want to use the working version, lets get the data
  if use_working_version and not version_number:
    # Get the schema and table info
    (schema, schema_table) = GetInfoSchemaAndTable(request, table)
    
    try:
      working_version = GetUserVersionWorkingRecord(request)

      working_data = utility.path.LoadYamlFromString(working_version['data_yaml'])

      #TODO(g): Also handle deletes
      pass
      
      # If we have the working data (!None), and this scheme ID is in it, then look deeper
      if working_data and schema['id'] in working_data:
        db_data = working_data[schema['id']]
        if schema_table['id'] in db_data:
          table_data = db_data[schema_table['id']]

          # If we have this record_id, in this table, in this database, then return the working record
          if record_id in table_data:
            # print '\n\nFound Working Record: %s\n\n' % table_data[record_id]
            
            # Update this data over the existing table data
            found_version_record = table_data[record_id]
            
            # Ensure it has a record ID.  We remove this from the data, since it doesnt change, and it needs to be added back on these transition points
            found_version_record['id'] = record_id
    
    except datasource.VersionNotFound, e:
      pass
  
  # Else, if they want to retrieve a specified version number
  elif version_number:
    # Look in the pending table first
    version_record = Get(request, 'version_pending', version_number)
    
    print 'Found pending version: %s' % version_record
    
    # If we didnt find it in pending, check in committed
    if not version_record:
      version_record = Get(request, 'version_commit', version_number)
      
      print 'Found committed version: %s' % version_record
    
      # If we didnt find it in commited, error
      if not version_record:
        raise RecordNotFound('Couldnt find version record: Table: %s:  Version: %s  Record ID: %s' % (table, version_number, record_id))
    
    
    # Get the schema and table info
    (schema, schema_table) = GetInfoSchemaAndTable(request, table)
    
    rollback_record_data = utility.path.LoadYamlFromString(version_record['rollback_data_yaml'], {})
    update_record_data = utility.path.LoadYamlFromString(version_record['data_yaml'], {})
    
    # Layer the rollback data (if any), underneath the update data
    record_data = rollback_record_data
    record_data.update(update_record_data)
    
    #TODO(g): Also handle deletes
    pass
    
    # If we have the working data (!None), and this scheme ID is in it, then look deeper
    if schema['id'] in record_data:
      db_data = record_data[schema['id']]
      if schema_table['id'] in db_data:
        table_data = db_data[schema_table['id']]

        # If we have this record_id, in this table, in this database, then return the working record
        if record_id in table_data:
          # Update this data over the existing table data
          found_version_record = table_data[record_id]
          
          # Ensure it has a record ID.  We remove this from the data, since it doesnt change, and it needs to be added back on these transition points
          found_version_record['id'] = record_id
    #TODO(g):REMOVE: Once we are sure this code is stable and we dont need it.  Lots of lines, and I think things are fine...
    #       
    #       print 'Found version record: %s' % found_version_record
    #     else:
    #       print 'Couldnt find record ID in version record: %s' % record_id
    #   else:
    #     print 'Couldnt find schema table ID in version record: %s' % schema_table['id']
    # else:
    #   print 'Couldnt find schema ID in version record: %s (%s)' % (schema['id'], type(schema['id']))
  
  
  #TODO(g): Confirm this is the primary key name, not just "id" all the time.  Can look this up in our schema_data_paths from connection_data...
  #TODO(g): Allow multiple fields for primary key, and do the right thing with them
  sql = "SELECT * FROM `%s` WHERE id = %s" % (table, int(record_id))
  result = connection.Query(sql)
  
  if result:
    record = result[0]
    
    # If we also found a version record, overlay it
    if found_version_record:
      record.update(found_version_record)
      # print 'Found Version Record, updating over Real Record:  Returning: %s (%s): %s' % (record_id, type(record_id), record)
    
  else:
    record = None
    
    # If we found a version record, use it.
    #TODO(g): In this case, every field should be present, or it will be a "corrupt" record.  How to test this?  Do I need to?  Maybe...  Optional not to allow this?  Not sure.  Thing about it...
    if found_version_record:
      record = found_version_record
   
      # print 'Couldnt find Real Record, but Found Version Record:  Returning: %s (%s): %s' % (record_id, type(record_id), record)
 
  return record


def Query(request, sql, params=None):
  """Perform a query without versioning.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    sql: string, SQL query to perform
    params: list of values, any value type will inserted into the query to be quoted properly
    version_number: int (default None), if an int, this is the version number in the version_change or version_commit
        tables.  version_change is scanned before version_commit, as these are more likely to be requested.
    use_working_version: boolean (default True), if True and version_number==None this will also look at any
        version_working data and return it instead the head table data, if it exists for this user.
  
  Returns: dict, single record key/values
  """
  # Get a connection
  connection = GetConnection(request)
  
  # Query
  result = connection.Query(sql, params)
  
  return result


def Filter(request, table, data=None, use_working_version=False, order_list=None, groupby_list=None, version_number=None):
  """Get 0 or more records from the datasource, based on filtering rules.  Works against a single table.
  """
  
  # print '\nFilter: %s: %s' % (table, data)
  
  if not data:
    data = {}
  
  base_sql = "SELECT * FROM `%s` WHERE %s %s %s"
  
  keys = data.keys()
  keys.sort()
  
  keys_ticked = []
  where_list = []
  values = []
  
  # Order By
  if order_list:
    order_by = ' ORDER BY %s' % ', '.join(order_list)
  else:
    order_by = ''
  
  # Group By
  if groupby_list:
    group_by = ' GROUP BY %s' % ', '.join(groupby_list)
  else:
    group_by = ''
  
  
  # Get our backticked wrapped insert keys, our value list, and our update setting
  for count in range(0, len(keys)):
    # Skip any fields that are NULL.  They are not helping us here, and cause problems with "=" vs "IS", because SQL implements NULL testing stupidly
    if data[keys[count]] == None:
      continue
    
    # Back tick column names
    ticked_key = '`%s`' % keys[count]
    keys_ticked.append(ticked_key)
    
    # If this is a normal value.  All non-normal tests should be wrapped in tuple (not other sequences) for the proper magic to occur.
    if type(data[keys[count]]) != tuple:
      # Update keys will reference the insert keys, so we dont have to specify the data twice (SQL does it)
      #TODO(g): Should I remove the primary key from this?  Not sure it's necessary.  Remove comment when proven it works without removing it (simpler)...
      where_list.append('%s = %%s' % ticked_key)
      
      # Values are passed in separate than the SQL string
      values.append(data[keys[count]])
    
    # Else, we want to do something being 
    else:
      #TODO(g): Do other op-codes too, so we can do many kinds of queries easily in this way
      pass
      
      # If the field is 'IN' a list of values
      if data[keys[count]][0].upper() == 'IN':
        match_list = data[keys[count]][1]
        where_in_str = '(%s)' % ', '.join(match_list)

        # Set the full statement here, which means we have to handle quoting the strings ourselves, if VARCHAR-like type
        where_list.append('%s IN %s' % (ticked_key, where_in_str))

      else:
        raise Exception('Filter: Unknown WHERE directive: %s' % data[keys[count]])
  
  
  # Build out strings to insert into our base_sql
  where_sql = ' AND '.join(where_list)
  
  # Create our final SQL, if we had WHERE list items
  if where_list:
    sql = base_sql % (table, where_sql, order_by, group_by)
  
  # Else, get all the records
  else:
    sql = "SELECT * FROM `%s` %s %s" % (table, order_by, group_by)
  
  # Log('\n\nGetFromData: %s: %s\nSQL:%s\nValues:%s\n' % (table, data, sql, values))
  
  
  # Get a connection
  connection = GetConnection(request)
  
  # Query
  rows = connection.Query(sql, values)
  
  
  # Assume we have no version data
  (update_version, delete_version) = (None, None)
  
  # If we want to use the working version, and we havent specified a version number (we dont want to mix both, too confusing.  Version Number is more explicit, it wins)
  #TODO(g): Move this section to generic_handler.py, because it can be generalized to all DB Handlers.
  if use_working_version and version_number == None:
    # Get the working version data for this user
    (update_version, delete_version) = GetWorkingVersionData(request)
    
    # print 'Version Record: Working: %s: \nUpdate: %s\nDelete: %s\n' % (is_pending, update_version, delete_version)


  # If we have specified an explicit version number, get it and see if it 
  if version_number:
    (version_record, is_pending) = GetInfoVersionNumber(request, version_number)
    update_version = utility.path.LoadYamlFromString(version_record['data_yaml'], {})
    delete_version = utility.path.LoadYamlFromString(version_record['delete_data_yaml'], {})
    
    print 'Version Record: Pending: %s: \nUpdate: %s\nDelete: %s\n' % (is_pending, update_version, delete_version)
  
  
  # If we have either update or deletes, from working, pending or committed versions.  We handle them all the same way.
  if update_version or delete_version:
    (schema, schema_table) = GetInfoSchemaAndTable(request, table)
    
    # Ensure rows is a list (mutable)
    if type(rows) == tuple:
      rows = list(rows)
    
    # Look to see if we have an Updates from our Working Version data, to make changes to the rows
    if schema['id'] in update_version:
      update_schema = update_version[schema['id']]
      
      if schema_table['id'] in update_schema:
        update_table = update_schema[schema_table['id']]
        
        # Get a list of all the row IDs, so we can see if we need to add any
        row_id_list = []
        
        # Loop over our row results
        for row in rows:
          # Add the row ID, so we know them all
          row_id_list.append(row['id'])
          
          # If the row we got from Filter() exists in our update_table, update those contents over the row
          if row['id'] in update_table:
            row.update(update_table[row['id']])
        
        # Loop over the update_table, and see if we have any entries we dont have in the rows, but that meet the requirement
        for (item_key, item) in update_table.items():
          # Set the ID field.  We remove it when putting it into the working table, because it doesnt change, so we have to add it back when creating records from that that table
          item['id'] = item_key
          
          # If this is a potential match
          if item['id'] not in row_id_list:
            
            # print 'Found potential match: %s' % item
            
            # Check if any of the filter key-values dont match, we only want to add it if they all match
            filter_data_matched = True
            for (filter_key, filter_value) in data.items():
              if filter_key not in item or item[filter_key] != filter_value:
                filter_data_matched = False
                # print '  Not matched: %s != %s' % (item.get(filter_key, '*KEY NOT FOUND*'), filter_value)
                break
            
            # If all the conditions are met
            if filter_data_matched:
              # Add this record to the rows.
              rows.append(item)
              
              #TODO(g): Sort these.  Order By, Group By, etc.
              pass
        
        #TODO(g): Order by, group by, etc.  We can control ALL the data so it's perfectly integrated, and looks like its part of the query
        pass
      
    # If we want these ordered, we need to sort them again
    if order_list:
      # Sort the rows by the order_list, so we have an ordered return set again
      rows = datasource.SortRows(rows, order_list)
    
    
    # print '\n\n+++ Delete version: %s' % delete_version
    
    # If we have any entries that we might need to delete, in our working version (delete versions)
    if schema['id'] in delete_version:
      delete_schema = delete_version[schema['id']]
      
      if schema_table['id'] in delete_schema:
        delete_table = delete_schema[schema_table['id']]
        
        # print '\n\n-*- Found entry in Delete Version table while in Filter: %s' % delete_table
        
        delete_rows = []
        
        # Add any rows matching our delete entry to the delete list
        for row in rows:
          if row['id'] in delete_table:
            # print 'Matched delete row: %s' % row
            delete_rows.append(row)
        
        # Remove any rows marked for deletion
        for row in delete_rows:
          # print 'Removing row: %s' % row
          rows.remove(row)



  return rows


def GetInfoVersionNumber(request, version_number):
  """Returns the version_* record information, for the specified version_number.
  
  First checks in version_pending, then checks version_commit.  It will only be in one or the other.
  
  Returns: tuple (dict (or None), boolean), if found, the (record, is_pending), else (None, is_pending)
  """
  record = Get(request, 'version_pending', version_number)
  is_pending = True
  
  # If it wasn't in pending, get it in committed (if it exists)
  if record == None:
    record = Get(request, 'version_commit', version_number)
    is_pending = False
  
  return (record, is_pending)


def GetWorkingVersionData(request, username=None):
  """Returns the version_working record's data_yaml, already parsed to Python data dict
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
  
  Returns: tuple of (dict, dict), which is (update_version, delete_version), respectively
  """
  # If one wasnt passed in, we use the requester
  if not username:
    username = request.username
  
  # Get the user_id
  user_list = Filter(request, 'user', {'name':username})
  
  if not user_list:
    return ({}, [])
  
  try:
    user = user_list[0]
    
    version_working = Get(request, 'version_working', user['id'])
    
    update_version = utility.path.LoadYamlFromString(version_working['data_yaml'])
    delete_version = utility.path.LoadYamlFromString(version_working['delete_data_yaml'])
  
    # Ensure they are properly initialized
    if update_version == None:
      update_version = {}
    if delete_version == None:
      delete_version = []
    
    return (update_version, delete_version)
  
  except Exception, e:
    print 'GetWorkingVersionData: Failed: %s: %s' % (username, e)
    return ({}, [])


def SetWorkingVersionData(request, working_version, delete_version=None):
  """Returns the version_working record's data_yaml, already parsed to Python data dict
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
  
  Returns: tuple of (dict, dict), which is (update_version, delete_version), respectively
  """
  user = GetUser(request)
  
  # Get the current working record directly, if we have it
  working_list = Filter(request, 'version_working', {'user_id': user['id']})
  if working_list:
    record = working_list[0]
  else:
    record = {'user_id': user['id'], 'delete_data_yaml': None}
  
  
  # Convert to YAML, for storage
  record['data_yaml'] = utility.path.DumpYamlAsString(working_version)
  
  # If we specified delete, update that too
  if delete_version != None:
    record['delete_data_yaml'] = utility.path.DumpYamlAsString(delete_version)
  
  # Update the working record
  SetDirect(request, 'version_working', record)


def Delete(request, table, record_id, noop=False, commit=True):
  """Delete a single record.
  
  NOTE(g): Processes single record deletes directly, sends fitlered deletes to DeleteFilter()

  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, primary key (ex: `id`) of the record in this table.  Use Filter() to use other field values
    noop: boolean (default False), if True do not actually query the database, (no operation)
    commit: boolean (default True), if True any queries that could be commited will be (single query transaction), if False then a later Commit() will be required
  
  Returns: None
  """
  # Get a connection
  connection = GetConnection(request)
  
  # Construct the query
  sql = "DELETE FROM `%s` WHERE id = %%s" % table
  
  # Delete the record
  if not noop:
    connection.Query(sql, [record_id], commit=commit)
  else:
    Log('Delete NO-OP: %s: %s' % (table, record_id))


def DeleteFilter(request, table, data, noop=False, commit=True):
  """Delete 0 or more records from the datasource, based on filtering rules.
  
  Can be a 'view', combining several lower level 'tables', which makes it a
  cascading delete.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, keys are strings (field names) and values are the values that must match (WHERE field=value)
    noop: boolean (default False), if True do not actually query the database, (no operation)
    commit: boolean (default True), if True any queries that could be commited will be (single query transaction), if False then a later Commit() will be required
  
  Returns: None
  """
  # Get a connection
  connection = GetConnection(request)
  
  # Ensure they arent trying to truncate
  if not data:
    raise InvalidArguments('DeleteFilter requires a dict with values to filter on.  Truncation of all data is not allowed from this function with an empty dictionary.')
  
  # INSERT values into a table, and if they already exist, perform an UPDATE on the fields
  base_sql = "DELETE FROM `%s` WHERE %%s" % table
  
  # Wrap all keys in backticks, so they cannot conflict with SQL keywords
  keys = data.keys()
  keys.sort()
  values = []
  where_format_list = []
  
  # Get our backticked wrapped insert keys, our value list, and our update setting
  for count in range(0, len(keys)):
    # Add the value (passed in as params list)
    values.append(data[keys[count]])
    
    # Back tick column names
    ticked_key = '`%s`' % keys[count]
    
    # Add the field's where clause, value will be added through param list
    where_format_list.append('%s = %%s' % ticked_key)
  
  # Generate where format string from our list, always AND for this filter.  More complex logic can be done directly through SQL, it's better than trying to wrap all options
  where_format = ' AND '.join(where_format_list)
  
  # Create our final SQL
  sql = base_sql % where_format
  
  # If we want to perform this operation (not no-op)
  if not noop:
    connection.Query(sql, values, commit=commit)
    
  else:
    Log('Delete Filter NO-OP: %s: %s' % (sql, values))




"""
Perform general actions against data for SchemaMan

We need to have short cuts to do things on our data, because we repeat access patterns too frequently in code.
These are things like looping over nested data inside of dictionaries, and ensuring nested dictinary elements exist, etc.

TODO(g):
  - Multiple iterations in a single for loop.  Go down multiple keys, and assign all the variables each time, like it has been flattened.  This will remove a lot of lines of code where I'm iterating over schema/table/records
"""


def EnsureNestedDictsExist(container, key_list, default_value, depth=0):
  """Iterates down a set of sub-containers (dicts), using key_list to navigate, and ensures all of them exist.
  
  default_value will be assigned to the final key, if it does not already exist.  This could be anything.
  
  Args:
    container: dict, parent container at current call depth
    key_list: list of keys (any dict acceptable hash type), will start from first to last, going deeper, making dicts as needed
    default_value: any, this is assigned into the final key dict's value, if that final key doesnt already exist
    depth: int, used for tracking recusion.  Do not manually set.
  
  Returns: None, changes made inside container, if any
  """
  # Return if we have nothing to do (no keys)
  if not key_list:
    return
  
  # Get the current key (first one), remove it from our list, get the remaining key count
  current_key = key_list[0]
  key_list.remove(current_key)
  remaining_keys = len(key_list)
  
  # If this is the last key
  if not remaining_keys:
    # If we dont have this current_key, then set the default value, and we are done
    if current_key not in container:
      container[current_key] = default_value
      return
  
  # Else, if there are more keys
  else:
    # If we dont have this current key, then we create a new dictionary at that key position, since we are all about nested dicts
    if current_key not in container:
      container[current_key] = {}
    
    # Now that we know we have a dictionary at our current key, and we have more keys, we recurse
    #   The new container will be this current key position, the key_list has current_key removed so it shrunk, and we pass the rest through, tracking recursion
    EnsureNestedDictsExist(container[current_key], key_list, default_value, depth=depth+1)


def NestedDictsToSingleDict(container, nested_depth, key_list=None, depth=0, final_data=None):
  """Takes a nested dictionary, and used key_depth (int) of the keys to create a tuple, which it will
  then store all the child keys into a single flat dict.
  
  Args:
    container: dict, nested dict (contains more dicts as values)
    nested_depth: int, number of nesteded dict depth of keys to capture for tuple in result single dict
    key_list: None or list, do not use.  Meant for tracking the current key list, as we recurse
    depth: int, do not use.  Meant for tracking recusion
    final_data: None of dict, do not use.  Meant for passing data along, to create the result
  
  Returns: dict, singlar flat dict, keyed on tuple, tuple is key_depth elements long
  """
  # This is what we will return, as our flat dict, keyed by nested dict keys in a tuple, with nested_depth length
  if final_data == None:
    final_data = {}
  
  # We start with an empty list, which we use for our tuple key
  if key_list == None:
    key_list = []
  
  
  # Process the current depth in our nested dicts
  for (cur_key, cur_value) in container.items():
    # Create a new list from our incoming key_list, and the new current key
    cur_key_list = key_list + [cur_key]
    
    # If we are at the nested_depth that we want to capture data at, assign the value into this tuple
    if len(cur_key_list) == nested_depth:
      final_data[tuple(cur_key_list)] = cur_value
    
    # Else, we want to go deeper into the nested dicts, so recurse again
    else:
      # Process this current value now, which is our next Nested Dict.  No return value, as we modify final_data in place (side effect)
      NestedDictsToSingleDict(cur_value, nested_depth, key_list=cur_key_list, depth=depth+1, final_data=final_data)
  
  return final_data


def GetFromNestedDict(container, key_list, default_value=None):
  """Returns the value at the nested dict starting from container, going down each layer using the key_list"""
  current_container = container
  
  # Process all our keys
  for key in key_list:
    # If we dont have this key, return the default
    if key not in current_container:
      return default_value
    
    # Update the current container by moving into this key
    current_container = current_container[key]
  
  # If we are at the end of our key_list, this is the value we were looking for, whatever it is
  return current_container




"""Data source handler.  Generic handler for all datasource types."""


import os
import threading
import time
from operator import itemgetter

from schemaman.utility.error import *
from schemaman.utility.path import *
import schemaman.utility as utility
from schemaman.utility.log import Log

# Schema Datasource functions
from request import Request
import tools

print tools

class RecordNotFound(Exception):
  """An expected record was not found."""


class VersionNotFound(Exception):
  """Version not found.  In version_* tables."""


def LoadConnectionSpec(path):
  """Load the connection specification."""
  if not os.path.isfile(path):
    Error('Connection path specified does not exist: %s' % path)
  
  try:
    data = LoadYaml(path)
    
    # Save our path, we will remove this if we save it
    data['__path'] = path
    
  except Exception, e:
    Error('Could not load connection spec YAML: %s: %s' % (path, e))
  
  return data


def SaveConnectionSpec(connection_data):
  """Load the connection specification."""
  try:
    data = dict(connection_data)
    path = data['__path']
    del data['__path']
    
    SaveYaml(path, data)
    
  except Exception, e:
    Error('Could not save connection spec YAML: %s: %s' % (path, e))
  
  return data


def DetermineHandlerModule(request):
  """Returns the handler module, which can handle these requests."""
  # If we didnt have a server_id specified, use the master_server_id
  if request.server_id == None:
    server_id = request.connection_data['datasource']['master_server_id']
  else:
    server_id = request.server_id
  
  # Find the master host, which we will assume we are connecting to for now
  found_server = None
  for server_data in request.connection_data['datasource']['servers']:
    if server_data['id'] == server_id:
      found_server = server_data
      break
  
  
  if not found_server:
    raise Exception('Could not find server ID in list of servers: %s' % server_id)
  
  
  # Test the host we found for it's connection type
  if found_server['type'] == 'mysql_56':
    import mysql_handler
    handler = mysql_handler
  
  else:
    raise Exception('Unknown Data Source Type: %s' % request.request['type'])
  
  
  # Add this handler to the request (if it doesnt have it), so it knows it needs to release these connections on Request object destruction
  request.AddHandler(handler)
  
  
  return handler


def TestConnection(request):
  """Connect to the datasource's database and ensure we can read from it."""
  handler = DetermineHandlerModule(request)
  
  result = handler.TestConnection(request)
  
  return result


def CreateSchema(request, schema):
  """Create a schema, based on a spec"""
  handler = DetermineHandlerModule(request)
  
  result = handler.CreateSchema(request, schema)
  
  return result


def ExtractSchema(request):
  """Export a schema, based on a spec, or everything"""
  handler = DetermineHandlerModule(request)
  
  result = handler.ExtractSchema(request)
  
  return result


def ExportSchema(request, path):
  """Export a schema, based on a spec, or everything"""
  handler = DetermineHandlerModule(request)
  
  handler.ExportSchema(request, path)
  
  return result


def UpdateSchema(request, schema):
  """Update a schema, based on a spec.
  
  Can go 'forward' or 'backwards' for version control, its still updating.
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.UpdateSchema(request, schema)
  
  return result


def ExportData(request, path):
  """Export/dump data from this datasource, based on spec, or everything"""
  handler = DetermineHandlerModule(request)
  
  result = handler.ExportData(request, path)
  
  return result


def ImportData(request, drop_first=False, transaction=False):
  """Import/load data to this datasource, based on spec, or everything.
  
  Args:
    drop_first: boolean, optional: If true, all data is dropped/deleted before
        the import occurs, otherwise it is an update.  Defaults to false to
        preserve data.
    transaction: boolean, optional: If true, import is done as a single
        transaction.  Defaults to False to avoid extra memory and slowness.
  
  Returns: None
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.ImportData(request, path)
  
  return result


def GetInfoSchema(request):
  """Returns the record for this schema data (schema)"""
  handler = DetermineHandlerModule(request)
  
  result = handler.GetInfoSchema(request)
  
  return result


def GetInfoSchemaTable(request, schema, table_name):
  """Returns the record for this schema table data (schema_table)"""
  handler = DetermineHandlerModule(request)
  
  result = handler.GetInfoSchemaTable(request, schema, table_name)
  
  return result


def GetInfoSchemaAndTable(request, table_name):
  """Returns the record for this schema table data (schema_table)
  
  This is a helper function, calls GetInfoSchema() and GetInfoSchemaTable()
  """
  schema = GetInfoSchema(request)
  
  schema_table = GetInfoSchemaTable(request, schema, table_name)
  
  return (schema, schema_table)


def GetInfoSchemaAndTableById(request, table_id):
  """Returns the record for this schema table data (schema_table)
  
  This is a helper function, calls GetInfoSchema() and GetInfoSchemaTable()
  """
  handler = DetermineHandlerModule(request)
  
  (schema, schema_table) = handler.GetInfoSchemaAndTableById(request, table_id)
  
  return (schema, schema_table)


def GetInfoSchemaTableField(request, schema_table, field_name):
  """Returns the record for this schema field data (schema_table_field)"""
  handler = DetermineHandlerModule(request)
  
  result = handler.GetInfoSchemaTableField(request, schema_table, field_name)
  
  return result


def GetUser(request, username=None, use_cache=True):
  """Returns user record (dict)"""
  handler = DetermineHandlerModule(request)
  
  # We allow an explicit username, but otherwise use the requester's username
  if not username:
    username = request.username
  
  
  #TODO(g): This needs to be configurable, so we can specify where to get authentication information...
  user = handler.GetUser(request, username, use_cache=use_cache)
  
  return user


def GetUserById(request, user_id, use_cache=True):
  """Returns user record (dict)"""
  handler = DetermineHandlerModule(request)
  
  #TODO(g): This needs to be configurable, so we can specify where to get authentication information...
  user = handler.GetUserById(request, user_id, use_cache=use_cache)
  
  return user


def RecordVersionsAvailable(request, table, record_id, username=None):
  """List all of the historical and currently available versions available for this record.
  
  Looks at 3 tables to figure this out: version_changelist_log (un-commited changes),
      version_commit_log (commited changes), version_working (single user changes)
  
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, primary key (ex: `id`) of the record in this table.  Use Filter() to use other field values
    username: string (default None), if not None, this is a specific user to check versions.  Otherwise the
        request.username is used.
    
  Returns: ...tbd...
  """
  handler = DetermineHandlerModule(request)
  
  # Get the user's ID
  user = GetUser(request, username)
  
  result = handler.RecordVersionsAvailable(request, table, record_id, user=user)
  
  return result


def CommitWorkingVersion(request):
  """Immediately takes everything in this user's Working Set and creates a change list.
  
  This is the short-cut for Change Management, so we don't need all the steps and still have versionining.
  
  See also: CreateChangeList() and CreateChangeListFromWorkingSet() and CommitWorkingVersionSingleRecord()
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.CommitWorkingVersion(request)
  
  return result


def CommitWorkingVersionSingleRecord(request, table, record_id):
  """TODO: Commit only a single record out of the working version.  Not all the working version records...
  
  Immediately takes a working version record and commits it, moving it through the rest of change management.
  
  This is the short-cut for Change Management, so we don't need all the steps and still have versionining.
  
  See also: CreateChangeList() and CreateChangeListFromWorkingSet() and CommitWorkingVersion()
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.CommitWorkingVersionSingleRecord(request, table, record_id)
  
  return result


def CreateChangeList(request, data):
  """Create a change list from the given table and record_id in the Working Set.
  
  This ensures we always have at least something in a change list, so we dont end up with empty ones where we dont know what
  went wrong.  We will always have some indication of what was going on if we find a change list.
  
  See also: CommitWorkingVersion() and CreateChangeListFromWorkingSet()
  
  Returns: int, version_number for this change list
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.CreateChangeList(request, record_id)
  
  return result


def CreateChangeListFromWorkingSet(request):
  """Create a Change List from the entire Working Set of version records.
  
  This is a fast way to prepare to commit everything that is currently being worked on.
  
  See also: CommitWorkingVersion() and CreateChangeList()
  
  Returns: int, version_number for this change list
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.CreateChangeListFromWorkingSet(request)
  
  return result


def AbandonWorkingVersion(request, table, record_id):
  """Abandon any current edits in version_working for this record.
  
  This does not effect Change Lists that are created, which must either be editted (removing record), or else
  the entire change list must be abandonded.
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.AbandonWorkingVersion(request, table, record_id)
  
  return result


def AbandonChangeList(request, change_list_id):
  """Abandon an open Change List, change_list_id==version_change.id
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.AbandonChangeList(request, change_list_id)
  
  return result


def Commit(request):
  """Commit a datasource transaction that is in the middle of a transaction."""
  handler = DetermineHandlerModule(request)
  
  result = handler.Commit(request)
  
  return result


def AbandonCommit(request):
  """Abandon Commit a datasource transaction that is in the middle of a transaction."""
  handler = DetermineHandlerModule(request)
  
  result = handler.AbandonCommit(request)
  
  return result


def Set(request, table, data, version_management=False, commit_version=False, version_number=None, commit=True):
  """Put (insert/update) data into this datasource.
  
  Works as a single transaction, if not using version_managament.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, record to set into table
    version_management: boolean (default False), if True this will use version management, if False will directly work with DB transactions
    commit_version: boolean (default False), if True, this will attempt to Commit the Version data after it has
        been stored in version_change as a single record update, without any additional VMCM actions
    version_number: int (default None), if an int, this is the version number in the version_change table to write to,
        if None this will create a new version_working entry
    commit: bool (default True), if not using version_management(==False) then this will be committed immediately,
        instead of waiting for a transactional Commit()
  
  Returns: int or None, if creating a new record this returns the newly created record primary key (ex: `id`), otherwise None
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.Set(request, table, data, version_management=version_management, commit_version=commit_version, version_number=version_number, commit=commit)
  
  return result


def SetDirect(request, table, data, commit=True):
  """Put (insert/update) data directly into this datasource.  Never use version management.
  
  Works as a single transaction, if not using version_managament.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, record to set into table
    commit: bool (default True), if not using version_management(==False) then this will be committed immediately,
        instead of waiting for a transactional Commit()
  
  Returns: int or None, if creating a new record this returns the newly created record primary key (ex: `id`), otherwise None
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.SetDirect(request, table, data, commit=commit)
  
  return result


def SetVersion(request, table, data, commit_version=False, version_number=None):
  """Put (insert/update) data into this datasource's Version Management tables (working, unless version_number is specified).
  
  This is the same as Set() except version_managament=True, which is more explicit.  This should be easier to read and type,
  and it clearly does something different.  By default Set() will 
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, record to set into table
    commit_version: boolean (default False), if True, this will attempt to Commit the Version data after it has
        been stored in version_change as a single record update, without any additional VMCM actions
    version_number: int (default None), if an int, this is the version number in the version_change table to write to,
        if None this will create a new version_working entry
  
  Returns: int or None, if creating a new record this returns the newly created record primary key (ex: `id`), otherwise None
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.SetVersion(request, table, data, commit_version=commit_version, version_number=version_number)
  
  return result


def SetFromUdnDict(request, data, version_number=None, use_working_version=True):
  """Set data from a dictionary with UDN keys (Universion Dotted Notation: 'table_name.row_id.field_name')
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    data: dict, record to set into table
    use_working_version: boolean (default True), if True and version_number==None this will also look at any
        version_working data and return it instead the head table data, if it exists for this user.
  """
  handler = DetermineHandlerModule(request)
  
  records = {}
  delete_records = []
  
  # Turn UDN into records
  for (udn, value) in data.items():
    # If we are instructed to ignore this one, skip it
    if udn.startswith('__skip.'):
      continue
    
    # Check if we are deleting a record
    is_delete = False
    if udn.startswith('__delete.'):
      is_delete = True
      
      # Strip off the delete, as we know about it now
      (_, udn) = udn.split('.', 1)
      
    
    (table, record_id, field) = udn.split('.')
    record_id = int(record_id)
    
    # Get the record from records, if it exists, otherwise create it
    record_key = (table, record_id)
    
    # Get the record from positive records or negative (delete)
    if not is_delete:
      # If we dont have an entry for this record yet
      if record_key not in records:
        records[record_key] = {'id': record_id}
      record = records[record_key]
      
      record[field] = value
    
    # Else, we are deleting, so use those delete records (as a list of tuples, since we only need to track table/record_id)
    else:
      # We want to store negative records in here too, just to get rid of any positive-update fields that may exist, so we nullify them.
      if record_key not in delete_records:
        delete_records.append(record_key)


  # Remove anything from our update records, if it also has a delete record.  Delete records win at this level, and neutralize updates.
  #   The reason Deletes win, is that Deletes only show up when someone has made an explicit call to delete a record, but positive update
  #   values will show up every request, because they are always present.  This is why deletes nullify positive updates.
  for record_key in delete_records:
    print 'Testing delete key: %s' % str(record_key)
    # If we have this record in positive-update records, then delete it, because the Delete records nullify it
    if record_key in records:
      print 'Deleting key: %s' % str(record_key)
      del records[record_key]
  
  
  import pprint
  print '\n\nSet UDN Records:\n%s\n' % pprint.pformat(records)
  print '\nSet UDN Delete Records:\n%s\n\n' % pprint.pformat(delete_records)
  
  # Set our data items
  for (record_key, record) in records.items():
    # Get the table from the record key.  We dont need the record_id, that is just for uniqueness in the key.  Set*() takes the key from the 'id' field
    (table, _) = record_key
    
    if not use_working_version:
      Set(request, table, record)
      
    else:
      SetVersion(request, table, record, version_number=version_number)
  
  
  # Delete our specified data items
  for (table, record_id) in delete_records:
    if not use_working_version:
      Delete(request, table, record_id)
      
    else:
      DeleteVersion(request, table, record_id, version_number=version_number)


def Get(request, table, record_id, version_number=None, use_working_version=True):
  """Get (select single record) from this datasource.
  
  Can be a 'view', combining several lower level 'tables'.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, primary key (ex: `id`) of the record in this table.  Use Filter() to use other field values
    version_number: int (default None), if an int, this is the version number in the version_change or version_commit
        tables.  version_change is scanned before version_commit, as these are more likely to be requested.
    use_working_version: boolean (default True), if True and version_number==None this will also look at any
        version_working data and return it instead the head table data, if it exists for this user.
    
  Returns: dict, single record key/values
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.Get(request, table, record_id, version_number=version_number, use_working_version=use_working_version)
  
  return result


def Query(request, sql, params=None):
  """Perform a query without versioning.
  
  Can be a 'view', combining several lower level 'tables'.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    sql: string, SQL query to perform
    params: list of values, any value type will inserted into the query to be quoted properly
    
  Returns: dict, single record key/values
  """
  handler = DetermineHandlerModule(request)
  
  result = handler.Query(request, sql, params=params)
  
  return result


def Filter(request, table, data=None, use_working_version=False, order_list=None, groupby_list=None, version_number=None):
  """Get 0 or more records from the datasource, based on filtering rules.
  
  Can be a 'view', combining several lower level 'tables'.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    data: dict, key is fields, value is the equality value.  Strict matching here.
    order_list: list of strings, fields to order by
    groupby_list: list of strings, fields to group by
    version_number: int (default None), if an int, this is the version number in the version_change or version_commit
        tables.  version_change is scanned before version_commit, as these are more likely to be requested.
    use_working_version: boolean (default True), if True and version_number==None this will also look at any
        version_working data and return it instead the head table data, if it exists for this user.
  """
  handler = DetermineHandlerModule(request)

  result = handler.Filter(request, table, data=data, order_list=order_list, groupby_list=groupby_list, version_number=version_number, use_working_version=use_working_version)
  
  return result


def GetWorkingVersionData(request, username=None):
  """Returns a dict or None, with the current working data (already parsed from `version_working.data_yaml`
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
  
  Returns: tuple of (dict, list), with labels (update_data, delete_data) 
  """
  handler = DetermineHandlerModule(request)

  result = handler.GetWorkingVersionData(request, username=username)
  
  return result


def GetWorkingVersionRecord(request, schema_id, schema_table_id, record_id, default_value=None, username=None):
  """Returns the Working Version record (usually dict), but we are not enforcing data types here.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    schema_id: int, schema.id
    schema_table_id: int, schema_table.id
    record_id: int, The record primary key `id` for the specified working version record
    default_value: any, will normally be None or passed in as {}
    username: string, optional, username to get Working Version for
  
  Returns: any, default_value if record does not existing in Working Version
  """
  (update_data, delete_data) = GetWorkingVersionData(request, username=username)
  
  # Assume our default_value, if we cannot find something better
  result = default_value
  
  # If we have this item in our working version data, then set it as our result value (overiding the default_value)
  if schema_id in update_data:
    if schema_table_id in update_data[schema_id]:
      if record_id in update_data[schema_id][schema_table_id]:
        result = update_data[schema_id][schema_table_id][record_id]
  
  return result


def SetWorkingVersionRecord(request, schema_id, schema_table_id, record_id, value):
  """Returns the Working Version record (usually dict), but we are not enforcing data types here.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    schema_id: int, schema.id
    schema_table_id: int, schema_table.id
    record_id: int, The record primary key `id` for the specified working version record
    value: dict
  
  Returns: None
  """
  (update_data, delete_data) = GetWorkingVersionData(request)
  
  # Ensure the schema exists in our update data
  if schema_id not in update_data:
    update_data[schema_id] = {}
  
  # Ensure the schema table exists in our update data
  if schema_table_id not in update_data[schema_id]:
    update_data[schema_id][schema_table_id] = {}
  
  # Set the record into the data
  update_data[schema_id][schema_table_id][record_id] = value
  
  # Save this data back into the working version data
  SetWorkingVersionData(request, update_data)


def SetWorkingVersionData(request, working_version, delete_version=None):
  """Returns a dict or None, with the current working data (already parsed from `version_working.data_yaml`
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
  
  Returns: dict or None
  """
  handler = DetermineHandlerModule(request)

  result = handler.SetWorkingVersionData(request, working_version, delete_version=delete_version)
  
  return result


def AcquireLock(request, lock, timeout=None, sleep_interval=0.1):
  """Spin loops until we can get this lock.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    lock: string, lock name.  Will be unique
    timeout: float (optional), time in seconds before timeout
    sleep_interval: float, time to sleep between checking on the lock
  
  Returns: boolean, did we get the lock?  True = yes. False = no.  This only matters if timeout is set, otherwise we will wait forever
  """
  Log('Acquire Lock: %s' % lock)
  
  done = False
  started = time.time()
  
  while not done:
    duration = time.time() - started
    
    # try:
    if 1:
      Set(request, 'schema_lock', {'name':lock})
      
      # It worked, we are done
      done = True
    
    # #TODO(g): Get the correct exception type here, so we only catch insertion failures
    # except Exception, e:
    #   print 'AcquireLock: Failed: %s: %s: %s' % (lock, duration, e)
    #   
    #   if timeout and duration > timeout:
    #     return False
    #   
    #   # Sleep for the specified time
    #   time.sleep(sleep_interval)
  
  return True


def ReleaseLock(request, lock):
  """Releases a lock.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    lock: string, lock name.  Will be unique
    
  Returns: boolean, did we release the lock?  True = yes. False = no.  If false, the lock was not set (which can indicate a problem)
  """
  Log('Release Lock: %s' % lock)
  
  lock_list = Filter(request, 'schema_lock', {'name': lock})
  
  if not lock_list:
    return False
  
  lock_record = lock_list[0]
  
  Delete(request, 'schema_lock', lock_record['id'])
  
  return True


def GetSchemaTableRowLockKey(request, table, record_id, schema=None):
  """
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
  
  Returns: str, lock key for specified table and record_id
  """
  # Ensure we have these.  We allow them to be passed in as a performance optimization
  if not schema:
    (schema, schema_table) = GetInfoSchema(request)
  
  lock = '%s.%s.%s' % (schema['id'], table, record_id)
  
  return lock


def GetNextNegativeNumber(request, table):
  """Returns the next negative number for a given table, from `schema_table.next_negative_id`
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
  """
  (schema, schema_table) = GetInfoSchemaAndTable(request, table)
  
  # Perform an internal lock, so we cant race on this
  #TODO
  
  # Get the lock key for this schema table row
  lock = GetSchemaTableRowLockKey(request, table, schema_table['name'], schema=schema)

  try:
    # Get the lock, so we dont collide on this
    AcquireLock(request, lock)
    
    # Get the next negative from the current storage
    next_negative_id = schema_table['next_negative_id']
    
    # Decrement the next negative ID, so we always get original ones, and they wont conflict
    schema_table['next_negative_id'] -= 1
    
    # Save with the new decremented number
    Set(request, 'schema_table', schema_table)
  
  finally:
    # Release the lock
    ReleaseLock(request, lock)
  
  return next_negative_id


def Delete(request, table, record_id):
  """Delete a single record.
  
  Use DeleteVersion() if you want version management over this delete.  This deleted the Real record.
  
  Use DeleteFilter() if you want to potentially delete many rows at once.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, primary key (ex: `id`) of the record in this table.  Use Filter() to use other field values
  """
  handler = DetermineHandlerModule(request)
  
  result =  handler.Delete(request, table, record_id)
    
  return result


def DeleteVersion(request, table, record_id, version_number=None):
  """Delete a single record from Working Version or a Pending Commit.
  
  Args:
    request: Request Object, the connection spec data and user and auth info, etc
    table: string, name of table to operate on
    record_id: int, primary key (ex: `id`) of the record in this table.  Use Filter() to use other field values
    version_number: int, this is the version number in the version_changelist.id
  
  Returns: None
  """
  if version_number != None:
    raise Exception('TBD: Version Number specific Deletes has not yet been implemented.')
  
  Log('Delete Version: %s: %s' % (table, record_id))
  
  (schema, schema_table) = GetInfoSchemaAndTable(request, table)
  
  user = GetUser(request)
  
  # Get the lock key for this schema table row
  lock = GetSchemaTableRowLockKey(request, table, record_id, schema=schema)
  
  try:
    # Acquire a lock, so we can work safely
    AcquireLock(request, lock)
    
    # Get the current working version
    version_working = Get(request, 'version_working', user['id'])
    
    # If we dont have a working version, make new dicts to store data in
    update_data = {}
    delete_data = {}
    
    # If, we have a working version, so get the data
    if version_working:
      update_data = utility.path.LoadYamlFromString(version_working['data_yaml'])
      delete_data = utility.path.LoadYamlFromString(version_working['delete_data_yaml'])
      
      if not update_data:
        update_data = {}
      
      if not delete_data:
        delete_data = {}

    
    # Check to see if this a Real record
    real_record = Get(request, table, record_id)
    
    # If we have a Real record, make an entry in the Delete Data, because we really want to delete this
    if real_record:
      
      # If we dont have the schema in our delete_data, add it
      if schema['id'] not in delete_data:
        delete_data[schema['id']] = {}
        
      # If we dont have the schema_table in our delete_data, add it
      if schema_table['id'] not in delete_data[schema['id']]:
        delete_data[schema['id']][schema_table['id']] = []
      
      # If we dont have this record in the proper place already (other records of that table to-delete), and this isnt a negative number, append it
      if record_id not in delete_data[schema['id']][schema_table['id']] and record_id >= 0:
        delete_data[schema['id']][schema_table['id']].append(record_id)
    
    # If we have an entry of this record in update_data, then remove that, because Delete always means to clear version data
    if schema['id'] in update_data:
      if schema_table['id'] in update_data[schema['id']]:
        if record_id in update_data[schema['id']][schema_table['id']]:
          # Delete the record from this update_data table, we are nulling that potential change
          del update_data[schema['id']][schema_table['id']][record_id]
    
    
    # Clean up the data, so we dont leave empty cruft around
    CleanEmptyVersionData(update_data)
    CleanEmptyVersionData(delete_data)
    
    # Add this to the working version record
    version_working['data_yaml'] = utility.path.DumpYamlAsString(update_data)
    version_working['delete_data_yaml'] = utility.path.DumpYamlAsString(delete_data)
    
    # Save the working version record
    Set(request, 'version_working', version_working)
  
  finally:
    # Ensure we release the lock
    ReleaseLock(request, lock)


def DeleteFilter(request, table, data, version_number=None, use_working_version=False):
  """Delete 0 or more records from the datasource, based on filtering rules.
  
  Can be a 'view', combining several lower level 'tables', which makes it a
  cascading delete.
  
  Works as a single transaction.
  
  NOTE(g): This is called by Delete(), and is not invoked from the CLI directly.
  """
  handler = DetermineHandlerModule(request)
  
  if version_number == None and use_working_version == False:
    result = handler.DeleteFilter(request, table, data)
    
  else:
    raise Exception('TBD: Havent implemented DeleteFilter() for versions yet')
  
  return result


def CommitVersionRecordToDatasource(request, version_commit_id, change_record, commit=True):
  """Pushes the contents of a version_commit row into the Real data source"""
  handler = DetermineHandlerModule(request)
  
  result = handler.CommitVersionRecordToDatasource(request, version_commit_id, change_record, commit=commit)
  
  return result


def CleanEmptyVersionData(version_data):
  """Cleans up any empty dicts or lists in the version_work data_yaml field data."""
  schema_keys = version_data.keys()
  
  # Loop over our schemas
  for schema_key in schema_keys:
    schema_data = version_data[schema_key]
    table_keys = schema_data.keys()
    
    for table_key in table_keys:
      table_data = schema_data[table_key]
      
      # If this is a dict type, we need to go one more level deep.  If not, we dont
      if type(table_data) == dict:
        record_keys = table_data.keys()
        
        # Loop over our records
        for record_key in record_keys:
          # If the record is empty, delete it
          if not table_data[record_key]:
            del table_data[record_key]
      
      # If the table is empty, delete it
      if not schema_data[table_key]:
        del schema_data[table_key]
    
    # If this schema is empty, delete it
    if not version_data[schema_key]:
      del version_data[schema_key]


def ListOfDictsToValueList(list_of_dicts, key, convert=None):
    """Returns a list of values, from the key of a list of dicts.  May convert type of values.

    Args:
        list_of_dicts: list of dicts
        key: value (any, for key in the dicts)
        convert: type, examples: str, int, float, etc.  Will be used as a function to convert value types.  Can also pass in any function that takes 1 value and emits 1 value.

    Returns: list of values
    """
    result_list = []

    for item in list_of_dicts:
        value = item[key]

        if convert:
          value = convert(value)

        result_list.append(value)

    return result_list


def ListOfDictsToDict(list_of_dicts, key):
    """Returns a dict, from the key of a list of dicts.

    Args:
        list_of_dicts: list of dicts
        key: value (any, for key in the dicts)

    Returns: list of values
    """
    result = {}

    for item in list_of_dicts:
        result[item[key]] = item

    return result


def ListOfDictsToKeyValueDict(list_of_dicts, key, value_key):
    """Returns a dict, from the key of a list of dicts.

    Args:
        list_of_dicts: list of dicts
        key: value (any, for key in the dicts)

    Returns: list of values
    """
    result = {}

    for item in list_of_dicts:
        result[item[key]] = item[value_key]

    return result


def SortRows(rows, order_list):
  """Returns a list of dicts.  Sorts rows by order_list, like Filter.
  
  Used for sorting Versioned data, which appends to Filter() real results, but must be re-ordered to be correct.
  
  Returns: list of dict (rows)
  """
  comparers = [
    ((i(col[1:].strip()), -1) if col.startswith('-') else (itemgetter(col.strip()), 1))
    for col in order_list
  ]
  
  def comparer(left, right):
    comparer_iter = (
      cmp(fn(left), fn(right)) * mult
      for fn, mult in comparers
    )
    return next((result for result in comparer_iter if result), 0)
    
  return sorted(rows, cmp=comparer)

  


"""
Interactive Input
"""

import readline


# Set up readline globally, so we dont need an init function for this to work
readline.parse_and_bind('tab: complete')
readline.parse_and_bind('set editing-mode vi')


def ReadLine(prompt):
  """Read a line, using readline controls."""
  line = raw_input(prompt)
  
  return line



def ReplaceUncleanChars(text, replacement_char='_'):
  unclean_chars = ' \t:;\'",.<>/?=+[]{}-`~!@#$%^&*()'
  
  for char in unclean_chars:
    text = text.replace(char, replacement_char)
  
  return text


def GetInputField(data, key, label, short_label, intro_string, force_lower=False, force_clean=False, force_strip=False, force_int=False, options=None):
  done = False

  while not done:
    # New input
    if key not in data:
      print '\n%s' % intro_string
      
      input_text = ReadLine('\nEnter %s: ' % short_label)
    
    # Else, we have existing input, prompting to continue or change
    else:
      print 'Hit enter (blank entry) if this is OK, or enter a new %s' % label
      print '\n%s: %s' % (label, data[key])
      input_text = ReadLine('\nEnter an updated %s, or enter to accept: ' % short_label)
    
    
    # If we got data, ingest it
    if input_text.strip():
      data[key] = input_text
      
      # Lower case
      if force_lower:
        data[key] = data[key].lower()
      
      # Strip
      if force_strip:
        data[key] = data[key].strip()
      
      # Clean characters
      if force_clean:
        data[key] = ReplaceUncleanChars(data[key])
      
      # Integer
      if force_int:
        try:
          data[key] = int(data[key])
        except ValueError, e:
          print '\nYou entered an invalid integer: %s\n' % data[key]
          
          del data[key]
      
      # Options
      if options and data[key] not in options:
        print '\nYou did not enter any of the possible options.\n'
        
        del data[key]
      
      # Alert them if we have changed their input
      if key in data and input_text != data[key]:
        print '\nThe %s has been re-written as this: %s\n' % (label, data[key])
    
    # Else, we have the data, and got empty input (newline = continue)
    elif key in data:
      # If we 
      if options and data[key] not in options:
        print 'The value must be one of these valid options: %s' % ', '.join(options)
      
      else:
        done = True


def CollectInitializationDataFromInput():
  """Collect information about this schema from the user interactively via stdin."""
  data = {}
  
  # Get the Schema Alias
  intro_string = 'Schema Names can be anything, and should be considered the "Fancy Name" with spaces and capitals, etc.\nSchema Alias should not have any spaces or capital letters, as they will be used as YAML file names, and in other places as variables.'
  GetInputField(data, 'alias', 'Schema Alias', 'alias', intro_string, force_lower=True, force_strip=True, force_clean=True)
  
  
  # Get the Schema Name
  intro_string = 'Schema Name should not have any spaces or capital letters, as they will be used as YAML file names, and in other places as variables.'
  GetInputField(data, 'name', 'Schema Name', 'name', intro_string, force_strip=True)
  
  
  # Get the Owner User
  intro_string = 'Owner User should be a username, or email, or whatever you use for authentication.  Leading and trailing white space is stripped.'
  GetInputField(data, 'owner_user', 'Owner User', 'user', intro_string, force_strip=True)
  
  
  # Get the Owner Group
  intro_string = 'Owner Group should be a username, or email, or whatever you use for authentication.  Leading and trailing white space is stripped.'
  GetInputField(data, 'owner_group', 'Owner Group', 'group', intro_string, force_strip=True)
  
  
  # Get the DB Host
  intro_string = 'Database Host is a valid hostname for a host that exists now and has a database on it.'
  GetInputField(data, 'database_host', 'Database Host', 'host', intro_string, force_strip=True)
  
  
  # Get the DB Port
  intro_string = 'Database Port is an integer which the database is running on.  Use 0 if there is no port.'
  GetInputField(data, 'database_port', 'Database Port', 'port', intro_string, force_strip=True, force_int=True)
  
  
  # Get the DB Name
  intro_string = 'Database name is the name of the database in the backend database.'
  GetInputField(data, 'database_name', 'Database Name', 'database name', intro_string, force_strip=True)
  
  
  # Get the DB User
  intro_string = 'Database user is a valid username to authenticate against the specified database.'
  GetInputField(data, 'database_user', 'Database User', 'user', intro_string, force_strip=True)
  
  
  # Get the DB Password
  intro_string = 'A path to a secret database password, stored in a text file on your local file system.'
  GetInputField(data, 'database_password_path', 'Database Password Path', 'password path', intro_string, force_strip=True)
  
  
  # Get the DB Type
  #TODO(g): Get these from our possible DB elements
  options = ['mysql_56']
  intro_string = 'Select the type of database backend from the list of options: %s' % ', '.join(options)
  GetInputField(data, 'database_type', 'Database Type', 'type', intro_string, force_strip=True, options=options)
  
  
  return data



"""
Process Actions
"""


import os
import pprint

# SchemaMan modules
import schemaman.utility
from schemaman.utility.log import Log
from schemaman.utility.error import *
from schemaman.utility.path import *
import schemaman.datasource as datasource

# Import action modules
import config
import populate
import test

# Mode for directories we create.
MODE_DIRECTORY = 0755

# Version of the connection_data spec
CONNECTION_SPEC_VERSION = 1


def ProcessAction(action, action_args, command_options):
  """Process the specified action, by it's action arguments.  Using command options."""
  
  #NOTE(g): This cannot be imported at the top, because it isnt ready yet, and it's a parent.  Python modules are terrible at pathing issues.
  import schemaman.action as action_module
  
  # If Action is info
  if action == 'info':
    if len(action_args) == 0:
      Usage('"init" action requires 1 argument: path schema definition YAML')
    elif not os.path.isfile(action_args[0]):
      Usage('"init" action requires arguments: %s: Is not a file' % action_args[0])
    
    schema_path = action_args[0]
    
    connection_data = datasource.LoadConnectionSpec(schema_path)
    
    print '\nConnection Specification:\n\n%s\n' % pprint.pformat(connection_data)
    print '\nTesting Connection:\n'
    
    request = datasource.Request(connection_data, 'testuser', 'testauth')
    
    # Attempt to connect to the DB to test it
    result = datasource.TestConnection(request)
    
    if result:
      print '\nConnection test: SUCCESS'
    else:
      print '\nConnection test: FAILURE'
  
  
  # If Action is action:  This is where we dump all kinds of functions, that dont need top-level access.  The long-tail of features.
  elif action == 'action':
    if len(action_args) < 3:
      Usage('All actions requires at least 3 arguments: <path schema definition YAML> <category> <action>  ...')
    elif not os.path.isfile(action_args[0]):
      Usage('All actions requires arguments: %s: Is not a file' % action_args[0])
    
    schema_path = action_args[0]
    
    connection_data = datasource.LoadConnectionSpec(schema_path)

    # Get all the args after the initial 3 args and use them as input for our Action function
    action_input_args = action_args[3:]
    
    #TODO(g): Turn this into YAML so that we can add into it.  Make sure it's multiple YAML files or something, so people can add their own without impacting the standard ones
    pass
    
    # -- Category --
    # Configure
    if action_args[1] == 'config':
      # Action
      if action_args[2] == 'version_change_management':
        if len(action_input_args) != 1:
          Error('action: populate: schema_into_db: Takes 1 argument: <path to target schema defintion YAML>')
        
        result = action_module.config.version_change_management.Action(connection_data, action_input_args)
        print result
      
      else:
        Usage('Unknown Action in Category: %s: %s' % (action_args[1], action_args[2]))
    
    # Populate
    elif action_args[1] == 'populate':
      # Action
      if action_args[2] == 'schema_into_db':
        if len(action_input_args) != 1:
          Error('action: populate: schema_into_db: Takes 1 argument: <path to target schema defintion YAML>')
        
        result = action_module.populate.schema_into_db.Action(connection_data, action_input_args)
        print result
      
      else:
        Usage('Unknown Action in Category: %s: %s' % (action_args[1], action_args[2]))
    
    # Test
    elif action_args[1] == 'test':
      # Action
      if action_args[2] == 'vmcm':
        # if len(action_input_args) != 1:
        #   Error('action: test: vmcm: Takes 1 argument: <path to target schema defintion YAML>')
        
        result = action_module.test.test_vmcm.Action(connection_data, action_input_args)
        print result
      
      else:
        Usage('Unknown Action in Category: %s: %s' % (action_args[1], action_args[2]))
    
    else:
      Usage('Unknown Category: %s' % action_args[1])
  
  
  # Else, Initialize a directory to be a SchemaMan location
  elif action == 'init':
    if len(action_args) == 0:
      Usage('"init" action requires 1 argument: directory to store schema definition inside of')
    elif not os.path.isdir(action_args[0]):
      Usage('"init" action requires arguments: %s: Is not a directory' % action_args[0])
    
    schema_dir = action_args[0]
    
    Log('Initializing Schema Definition Directory: %s' % schema_dir)
    
    # Collect the initialization data from the user
    data = utility.interactive_input.CollectInitializationDataFromInput()
    
    schema_path = '%s/%s.yaml' % (schema_dir, data['alias'])
    
    # Check to see if we havent already created this schema definition.  We don't allow init twice, let them clean it up.
    if os.path.exists(schema_path):
      Error('The file you requested to initialize already exists, choose another Schema Alias: %s' % schema_path)
  
    # Create the location for our schema record paths.  This is the specific data source record schema, whereas the schema_path is the definition for the data set
    schema_record_path = '%s/schema/%s.yaml' % (schema_dir, data['alias'])
    
    
    # We need to create this directory, if it doesnt exist
    schema_record_path_dir = os.path.dirname(schema_record_path)
    if not os.path.isdir(schema_record_path_dir):
      os.makedirs(schema_record_path_dir, mode=MODE_DIRECTORY)
    
    
    # If we dont have this file yet, create it, so we can write into it
    if not os.path.isfile(schema_record_path):
      with open(schema_record_path, 'w') as fp:
        # Write an empty dictionary for YAML
        fp.write('{}\n')
    
    
    # Format the data into the expandable schema format (we accept lots more data, but on init we just want 1 set of data)
    schema_data = {
      'alias': data['alias'],
      'name': data['name'],
      'owner_user': data['owner_user'],
      'owner_group': data['owner_group'],
      'datasource': {
        'database': data['database_name'],
        'user': data['database_user'],
        'password_path': data['database_password_path'],
        'master_server_id': 1,
        'servers': [
          {
            'id': 1,
            'host': data['database_host'],
            'port': data['database_port'],
            'type': data['database_type'],
          },
        ],
      },
      'schema_paths': [schema_record_path],
      'value_type_path': 'data/schema/value_types/standard.yaml',
      'actions': {},
      'version': CONNECTION_SPEC_VERSION,
    }
    
    SaveYaml(schema_path, schema_data)
    
    Log('Initialized new schema path: %s' % schema_path)
  
  
  # Else, if Action prefix is Schema
  elif action == 'schema':
    # Ensure there are sub-actions, as they are always required
    if len(action_args) == 0:
      Usage('"schema" action requires arguments: create, export, extract, migrate, update')
    
    
    elif action_args[0] == 'create':
      connection_data = datasource.LoadConnectionSpec(action_args[1])
      
      result = datasource.CreateSchema(connection_data)
    
    
    # Export the current DB schema to a specified data source
    elif action_args[0] == 'export':
      if len(action_args) == 3:
        Usage('"schema export" action requires arguments: <path to connection spec> <path to export data to>')
      
      connection_data = datasource.LoadConnectionSpec(action_args[1])
      
      target_path = action_args[2]
      
      if not os.path.isdir(os.path.dirname(target_path)):
        Usage('Path specified does not have a valid directory: %s' % target_path)
      
      result = datasource.ExportSchema(connection_data, target_path)
    
    
    # Extract is the opposite of "update", and will get our DB schema and put it into our files where we "update" from
    elif action_args[0] == 'extract':
      if len(action_args) == 1:
        Usage('"schema extract" action requires arguments: <path to connection spec>')
      
      connection_data = datasource.LoadConnectionSpec(action_args[1])
      
      result = datasource.ExtractSchema(connection_data)
      
      print 'Extract Schema:'
      
      pprint.pprint(result)
      
      # Save the extracted schema to the last file in the connection data (which updates over the rest)
      #TODO(g): We should load all the previous files, and then only update things that are already in the current files, and save them, and then add any new things to the last file.
      output_path = connection_data['schema_paths'][-1]
      
      SaveYaml(output_path, result)
    
    
    elif action_args[0] == 'migrate':
      # Export from one, and import to another, in one step
      source_result = datasource.ExportSchema()
      target_result = datasource.UpdateSchema(source_result)
    
    
    # Update the database based on our schema spec files
    elif action_args[0] == 'update':
      connection_data = datasource.LoadConnectionSpec(action_args[1])
      
      result = datasource.UpdateSchema(connection_data)
    
    
    # ERROR
    else:
      Usage('Unknown Schema action: %s' % action)
  
  
  # Else, if Action prefix is Data
  elif action == 'data':
    if len(action_args) == 0:
      Usage('"data" action requires arguments: export, import')
    elif action_args[0] == 'export':
      result = datasource.ExportData()
    
    elif action_args[0] == 'import':
      result = datasource.ImportData()
    
    # ERROR
    else:
      Usage('Unknown Data action: %s' % action)
  
  # Put
  elif action == 'put':
    result = datasource.Put()
  
  # Get
  elif action == 'get':
    result = datasource.Get()
  
  # Filter
  elif action == 'filter':
    result = datasource.Filter()
  
  # Delete
  elif action == 'delete':
    result = datasource.Delete()
  
  # ERROR
  else:
    Usage('Unknown action: %s' % action)



"""
Request object to track user requests against our data sources.
"""


import threading


import generic_handler


# Every time we start, we use our initial global request counter, and on the beginning of a request, we increment it,
#   so that we can keep transactions together if they use the same request number.  Basically a singleton.
GLOBAL_REQUEST_COUNTER = 1
GLOBAL_REQUEST_COUNTER_LOCK = threading.Lock()


# Defaults
DEFAULT_AUTO_COMMIT = True


class RequestInvalid(Exception):
  """This request is no longer valid."""


class Request:
  """Contains request information, and can close transactions due to scope GC collections."""
  
  def __init__(self, connection_data, username, authentication, request_number=None, server_id=None, use_version_management=True, auto_commit=DEFAULT_AUTO_COMMIT):
    self.connection_data = connection_data
    
    if username == None:
      raise RequestInvalid('A username must be specified: %s: %s' % (connection_data, username))
    
    self.username = username
    self.authentication = authentication
    
    # Ensure we have a request number
    if not request_number:
      self.request_number = self.GetRequestNumber()
    
    else:
      self.request_number = request_number
    
    
    # If we have specified a specified server in a Connection Spec datasources, we can have it set here
    self.server_id = server_id
    
    # If True (default), then will use version management features, otherwise we just directly write to the Head of the datasource.
    #   Writing directly to head is necessary for all internal operations, which we do not want to bog down our system with
    #   automaticg version management rollback stuff.
    self.use_version_management = use_version_management
    
    # If True, we will auto commit any versions...
    self.auto_commit = auto_commit
    
    # Keep track of all handlers we use, so we can release the connections when we are done
    #TODO(g): Add handlers as we use them to this request, so we know what we need to release and dont have to try to scan the world...  Specific work is more efficient than scan-the-world work...
    self.datasource_handlers = []
    
    # Log is a list of tuples (text, data), 
    self.log = []
    
    # Track whether we have released this yet
    self.is_released = False
    
    
    # Get the user record
    #TODO(g): Where we get the user records needs to be configurable, and currently isnt.  Fix later, same with VMCM
    self.user = generic_handler.GetUser(self, self.username)
  
  
  def __del__(self):
    """Going out of scope, ensure we release any connections that we have."""
    try:
      self.Release()
    
    # Probably the interpretter is shutting down if we are getting this ("'NoneType' object is not callable"), so ignore it
    except TypeError, e:
      pass
    
  
  def Release(self):
    """Cleanup this request.  Relase it."""
    # We are released, so we should not be used anymore
    self.is_released = True
    
    # Release all our connections
    self.ReleaseConnections()

  
  def Log(self, text, data):
    """Log any data we want to about this request."""
    self.log.append((text, data))
  
  
  def ReleaseConnections(self):
    """Release any connections we have open and tied to this request_number (wont close them)"""
    for handler in self.datasource_handlers:
      handler.ReleaseConnections(self)


  def AddHandler(self, handler):
    """Allows us to track all handlers that might have datasource connections for us, so we can release them explicitly and efficiently."""
    if self.is_released:
      raise RequestInvalid('This request has already been released.  It cannot be used further.')
    
    if handler not in self.datasource_handlers:
      self.datasource_handlers.append(handler)


  def GetRequestNumber(self):
    """Returns an int, the next available request number.
    
    This number can be used with any number of data sets simultaneously, as it is globally unique, and
    each database connection will reside in it's own server's pool, so any number of databases
    can be queried with the same request_number.
    """
    global GLOBAL_REQUEST_COUNTER
    
    # Thread safe
    try:
      GLOBAL_REQUEST_COUNTER_LOCK.acquire()
      
      request_number = GLOBAL_REQUEST_COUNTER
      GLOBAL_REQUEST_COUNTER += 1
    
    finally:
      GLOBAL_REQUEST_COUNTER_LOCK.release()
    
    return GLOBAL_REQUEST_COUNTER
  



"""
Format data for various data source operations.  Pre-query or post-query.
"""


def FormatStringOrNullFromDict(data, key, strict=False):
  """Returns a quoted and escaped string from the data[key] value, or NULL to insert into a field column
  
  If strict=True, then an exception is thrown if the key is missing, otherwise it is NULL.
  """
  if key not in data and strict:
    raise Exception('Missing data key: %s' % key)
  
  if key in data and data[key] != None:
    output = "'%s'" % SanitizeSQL(str(data[key]))
  else:
    output = 'NULL'
  
  return output


def FormatIntOrNullFromDict(data, key, strict=False):
  """Returns a quoted and escaped string from the data[key] value, or NULL to insert into a field column
  
  If strict=True, then an exception is thrown if the key is missing, otherwise it is NULL.
  """
  if key not in data and strict:
    raise Exception('Missing data key: %s' % key)
  
  if key in data and data[key] != None:
    output = "%s" % int(data[key])
  else:
    output = 'NULL'
  
  return output


def ConvertListToDict(list_data, key_field, multi_key_separtor='.'):
  """Returns a dict keyed on key_field of list_data item dict data.  list_data is a list of dicts.
  
  key_field can be string (single value) or list (combined value into a key.field.value dict key)
  """
  data = {}
  
  for item in list_data:
    # If this is a single key field, then 
    if type(key_field) != list:
      data[item[key_field]] = item
      
    else:
      key = multi_key_separator.join(key_field)
      data[key] = item
  
  return data


def ConvertListToDictList(list_data, key_field):
  """Returns a dict of lists of dicts keyed on key_field of list_data item dict data.  list_data is a list of dicts.
  
  key_field can be string (single value) or list (combined value into a key.field.value dict key)
  """
  data = {}
  
  for item in list_data:
    # If this is a single key field, then 
    if type(key_field) != list:
      if item[key_field] not in data:
        data[item[key_field]] = []
      
      data[item[key_field]].append(item)
      
    else:
      key = multi_key_separator.join(key_field)
      if key not in data:
        data[key] = []
      
      data[key].append(item)
  
  return data




"""
Logging
"""


import logging
import time
import sys


# Default logging level
LOG_LEVEL_DEFAULT = logging.INFO


# Minimum log level we will actually log.  Beneath this we ignore (can ignore DEBUG, WARN, etc)
# LOG_LEVEL_MINIMUM = logging.INFO
LOG_LEVEL_MINIMUM = logging.DEBUG


def Log(text, log_level=None):
  """Send log messages to STDERR, so we can template to STDOUT by default (no output path, easier testing)
  
  Implement with rotating files using Python logging, prepped for that.  De-duping, counting, mini-stack (calling function, lineno) can be added, etc.
  """
  # Dont log this if it is beneath our default level, and it was explicitly passed in.
  if log_level and log_level < LOG_LEVEL_MINIMUM:
    return
  
  elif log_level == None:
    log_level = LOG_LEVEL_DEFAULT
  
  
  timestamp = '[%d-%02d-%02d %02d:%02d:%02d] ' % time.localtime()[:6]
  sys.stderr.write(timestamp + str(text) + '\n')
  sys.stderr.flush()




"""
Actions: Populate: Schema Into DB

Import a schema into a set of database tables:  database, table, field
"""



# SchemaMan libraries
import schemaman.utility as utility
from schemaman.utility.interactive_input import *
import schemaman.datasource as datasource


# This action's command on the CLI and also in the connection_data.actions dict as a key for our data
ACTION = 'populate__schema_into_db'


def Action(connection_data, action_input_args):
  """Perform action: Populate schema into DB"""
  print 'Populate Schema Into DB'
  
  data = CollectData(connection_data, action_input_args)
  
  # If we got invalid data or aborted, quit this action
  if data == None:
    return '\nAborting...'


  # Load all the schemas, and overlay them to get the sum of all of them (in order, so deterministic, last write wins)
  total_schema = {}
  for schema_path in connection_data['schema_paths']:
    schema = utility.path.LoadYaml(schema_path)
    
    total_schema.update(schema)
  
  
  # Put these tables for this database into the target schema
  target_connection_data = datasource.LoadConnectionSpec(action_input_args[0])
  
  
  database_name = connection_data['datasource']['database']
  
  print 'Populating Schema for Database: %s' % database_name
  
  
  # Get our request number.  We can use this with all data sources.
  target_request = datasource.Request(target_connection_data, 'username', 'auth')
  
  
  # Get the existing entry for this, if it exists
  schema_database_list = datasource.Filter(target_request, data['table_database'], {'name':data['table_database']})
  
  print 'Filter: Schema DB: %s' % schema_database_list
  
  # If we dont already have this record
  if not schema_database_list:
    # Create it, and get it's record again so we have it's ID
    schema_record = {'name':data['table_database'], 'mysql_user':connection_data['datasource']['user'],
                     'mysql_password_path':connection_data['datasource']['password_path'],
                     'mysql_hostname':connection_data['datasource']['servers'][0]['host']}
    schema_database_id = datasource.Set(target_request, data['table_database'], schema_record)
    
    print 'Created record: %s' % schema_database_id
    
    # Get the record we just put in
    schema_database = datasource.Get(target_request, data['table_database'], schema_database_id)

    print 'Fetched record: %s' % schema_database
  
  else:
    schema_database = schema_database_list[0]
  
  
  # Loop over all our tables
  for (table, table_data) in schema.items():
    print 'Populating Schema for Table: %s' % table

    table_record = {'name':table, 'schema_id':schema_database['id']}
    schema_table_list = datasource.Filter(target_request, data['table_table'], table_record)
    
    # If we dont have this table, create it
    if not schema_table_list:
      schema_table_id = datasource.Set(target_request, data['table_table'], table_record)
    
    else:
      schema_table_id = schema_table_list[0]['id']
    
    for (field, field_data) in table_data.items():
      print 'Populating Schema for Field: %s: %s' % (table, field)
    
      field_record = {'name':field, 'schema_table_id':schema_table_id}
      schema_field_list = datasource.Filter(target_request, data['table_field'], field_record)
      
      # If we dont have this table, create it
      if not schema_field_list:
        # Determine the Value Type ID
        #TODO(g): This isnt always part of the process...  We need some other way to do this update, and a post-script or something...?
        if field_data['type'] == 'int':
          value_type_id = 2
        elif field_data['type'] == 'varchar':
          value_type_id = 1
        elif field_data['type'] == 'text':
          value_type_id = 1
        else:
          raise Exception('Unknown value type: %s' % field_data['type'])
        
        if field.endswith('data_json'):
          value_type_id = 11
        
        # We want to set all the data, we only wanted to search and table/fieldname
        field_record = {'name':field, 'schema_table_id':schema_table_id, 'is_primary_key':field_data['pkey'],
                        'allow_null':field_data['allow_null'], 'default_value':field_data['default'], 'value_type_id':value_type_id}
        schema_field_id = datasource.Set(target_request, data['table_field'], field_record)


def CollectData(connection_data, action_input_args):
  """Collect data that we need, if we dont already have it stored in our connection_data.actions dict"""
  data = {}
  
  # If we have data in our connection_data, use it
  if ACTION in connection_data['actions']:
    #TODO(g): Validate that this data is correct
    data = connection_data['actions'][ACTION]
  
    print '\nThese are the tables you specified:'
    print '  - Database: %s' % data['table_database']
    print '  - Table: %s' % data['table_table']
    print '  - Field: %s' % data['table_field']
    print
    print 'Is this correct?  (Enter Yes to use these values, anything else to enter new values)'
    
    confirmation = ReadLine('$ ')
    
    # Use the saved data, otherwise collect new data
    if confirmation.lower() == 'yes':
      return data
    else:
      data = {}
  
  
  # Get the Database table name
  intro_string = 'Name of table to put the Database in.'
  GetInputField(data, 'table_database', 'Database Table', 'database table', intro_string, force_strip=True)
  
  # Get the Table table name
  intro_string = 'Name of table to put the Tables in.'
  GetInputField(data, 'table_table', 'Table Table', 'table table', intro_string, force_strip=True)
  
  # Get the Field table name
  intro_string = 'Name of table to put the Fields in.'
  GetInputField(data, 'table_field', 'Field Table', 'field table', intro_string, force_strip=True)

  
  print '\nThese are the tables you specified:'
  print '  - Database: %s' % data['table_database']
  print '  - Table: %s' % data['table_table']
  print '  - Field: %s' % data['table_field']
  print
  print 'Is this correct?  (Enter Yes, Save or this action will abort)'
  
  confirmation = ReadLine('$ ')
  
  # Aborting
  if confirmation.lower() not in ('yes', 'save'):
    return None
  
  
  # Save our connection data with the updated information
  if confirmation.lower() == 'save':
    connection_data['actions'][ACTION] = data
    
    datasource.SaveConnectionSpec(connection_data)
  
  
  return data
  
  


from setuptools import setup, find_packages

setup(name='schemaman',
      version='0.1.40',
      description='SchemaMan is a schema manager and query wrapper for multiple database backends, with Version and Change Management.',
      url='http://github.com/ghowland/schemaman',
      author='Geoff Howland',
      author_email='geoff@gmail.com',
      license='MIT',
      packages=['schemaman'],
      include_package_data=True,
      data_files=[('', ['LICENSE', 'README.md'])],
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Topic :: Database :: Front-Ends',
      ],
      entry_points = {
        'console_scripts': ['schemaman=schemaman.schemaman:Main'],
      },
      zip_safe=False)




"""
File path related functions.

Standard operations are wrapped here to be shorter and also to provide a place to create consistency between our operations.
"""


import yaml
import collections
import StringIO

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper as Dumper


# Key is file path, this is to store YAML decoded data, for performance
DATA_CACHE = {}


class SafeDumper(Dumper):
  """Create a class to handle unicode messiness in YAML dumps."""
  
  def ignore_aliases(self, data):
    """
    A dumper that will never emit aliases.
    """
    return True


# Strings as strings
SafeDumper.add_representer(unicode, yaml.representer.SafeRepresenter.represent_str)

# Unicode as strings
SafeDumper.add_representer(unicode, yaml.representer.SafeRepresenter.represent_unicode)


def LoadYaml(path, use_cache=True):
  """Load data from YAML format."""
  global DATA_CACHE
  
  if not use_cache or path not in DATA_CACHE:
    DATA_CACHE[path] = yaml.load(open(path))
  
  return DATA_CACHE[path]


def SaveYaml(path, data):
  """Save data in YAML format."""
  yaml.safe_dump(data, open(path, 'w'))


def LoadYamlFromString(text, default_value=None, strict=False):
  """Loads YAML data from a string
  
  Returns: any data container type storable in YAML, or None
  """
  # None is never a failure, its just the default value
  if text == None:
    return default_value
  
  # If we are not strict, we will handle all exceptions as setting default value
  if not strict:
    try:
      result = yaml.load(StringIO.StringIO(text))
    
    except Exception, e:
      result = default_value
  
  # Else, we will raise exceptions as normal.  Separate else section so stack trace is from the failure point, and not hidden
  else:
      result = yaml.load(StringIO.StringIO(text))
  
  return result


def DumpYamlAsString(data, deep_unicode_conversion=False):
  """Wraps dumping YAML, in the way we need it to diff against existing YAML."""
  # Convert all unicode to strings, to get rid of annoying YAML formatting
  if deep_unicode_conversion:
    data = DeepUnicodeToString(data)
  
  # Better than safe_dump, for clarity.
  result = yaml.dump(data, Dumper=SafeDumper)
  
  return result


def DeepUnicodeToString(data):
  """Converts deep data structures into all strings.  No more unicode."""
  if type(data) is unicode:
    return str(data.encode('ascii', 'ignore'))
  if type(data) is list:
    return [DeepUnicodeToString(a) for a in data]
  if type(data) is dict:
    return dict((DeepUnicodeToString(key), DeepUnicodeToString(value)) for key, value in data.iteritems())
  if isinstance(data, collections.OrderedDict):
    return collections.OrderedDict((DeepUnicodeToString(key), DeepUnicodeToString(value)) for key, value in data.iteritems())
  
  return data

